###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

import json
import pathlib
import random
import re
import string

# Requirements
import httpx
import rich

import caterva2 as cat2

# Project
from caterva2 import api_utils, utils


def handle_errors(func):
    def wrapper(*args):
        try:
            func(*args)
        except httpx.HTTPStatusError as error:
            response = error.response
            try:
                error = response.json()["detail"]
            except json.decoder.JSONDecodeError:
                error = response.text
            print("Error:", error)

    return wrapper


def dataset_with_slice(path):
    match = re.match("(.*)\\[(.*)]", path)
    if match is None:
        params = {}
    else:
        path, slice_ = match.groups()
        params = {"slice_": slice_}

    return pathlib.Path(path), params


@handle_errors
def cmd_roots(client, args):
    data = client.get_roots()
    if args.json:
        print(json.dumps(data))
        return

    for name, root in data.items():
        if root["subscribed"] is True:
            print(f"{name} (subscribed)")
        else:
            print(name)


@handle_errors
def cmd_subscribe(client, args):
    data = client.subscribe(args.root)
    if args.json:
        print(json.dumps(data))
        return

    print(data)


@handle_errors
def cmd_list(client, args):
    data = client.get_list(args.root)
    if args.json:
        print(json.dumps(data))
        return

    for item in data:
        print(f"{item}")


@handle_errors
def cmd_url(client, args):
    data = api_utils.get_download_url(args.dataset, args.urlbase)
    if args.json:
        print(json.dumps(data))
        return
    print(data)


@handle_errors
def cmd_info(client, args):
    print(f"Getting info for {args.dataset}")
    data = client.get_info(args.dataset)

    # Print
    if args.json:
        print(json.dumps(data))
        return
    rich.print(data)


@handle_errors
def cmd_show(client, args):
    path, params = args.dataset
    slice_ = params.get("slice_", None)
    data = client.fetch(path, slice_=slice_)

    # Display
    if isinstance(data, bytes):
        try:
            print(data.decode())
        except UnicodeDecodeError:
            print("Binary data")
    else:
        print(data)
        # TODO: make rich optional in command line
        # rich.print(data)


@handle_errors
def cmd_move(client, args):
    moved = client.move(args.dataset, args.dest)
    print(f"Dataset {args.dataset} moved to {moved}")


@handle_errors
def cmd_copy(client, args):
    copied = client.copy(args.dataset, args.dest)
    print(f"Dataset {args.dataset} copied to {copied}")


@handle_errors
def cmd_download(client, args):
    path = client.download(args.dataset, args.localpath)
    print(f"Dataset saved to {path}")


@handle_errors
def cmd_upload(client, args):
    path = client.upload(args.localpath, args.dataset)
    print(f"Dataset stored in {path}")


@handle_errors
def cmd_remove(client, args):
    removed = client.remove(args.dataset)
    print(f"Dataset (or directory contents) removed: {removed}")


@handle_errors
def cmd_adduser(client, args):
    newpass = args.newpass or "".join(random.choice(string.ascii_letters) for _ in range(8))
    message = client.adduser(
        args.newuser,
        newpass,
        args.superuser,
    )
    print(message)


@handle_errors
def cmd_deluser(client, args):
    message = client.deluser(args.user)
    print(message)


@handle_errors
def cmd_listusers(client, args):
    data = client.listusers(args.user)
    if args.json:
        print(json.dumps(data))
        return

    for user in data:
        print(user)


def main():
    conf = utils.get_conf()
    parser = utils.get_parser()
    parser.add_argument(
        "--subscriber",
        dest="urlbase",
        type=utils.urlbase_type,
        default=conf.get("subscriber.url", cat2.sub_urlbase_default),
    )
    parser.add_argument("--username", default=conf.get("client.username"))
    parser.add_argument("--password", default=conf.get("client.password"))
    subparsers = parser.add_subparsers(required=True)

    # roots
    help = "List all the available roots in a broker."
    subparser = subparsers.add_parser("roots", help=help)
    subparser.add_argument("--json", action="store_true")
    subparser.set_defaults(func=cmd_roots)

    # subscribe
    help = "Request access to the datasets in a root."
    subparser = subparsers.add_parser("subscribe", help=help)
    subparser.add_argument("--json", action="store_true")
    subparser.add_argument("root")
    subparser.set_defaults(func=cmd_subscribe)

    # list
    help = "List all the available datasets in a root. Needs to be subscribed to the root."
    subparser = subparsers.add_parser("list", aliases=["ls"], help=help)
    subparser.add_argument("--json", action="store_true")
    subparser.add_argument("root")
    subparser.set_defaults(func=cmd_list)

    # copy
    help = "Copy a dataset to a different root."
    subparser = subparsers.add_parser("copy", aliases=["cp"], help=help)
    subparser.add_argument("dataset", type=pathlib.Path)
    subparser.add_argument("dest")
    subparser.set_defaults(func=cmd_copy)

    # move
    help = "Move a dataset to a different root."
    subparser = subparsers.add_parser("move", aliases=["mv"], help=help)
    subparser.add_argument("dataset", type=pathlib.Path)
    subparser.add_argument("dest")
    subparser.set_defaults(func=cmd_move)

    # remove
    help = "Remove a dataset from the subscriber."
    subparser = subparsers.add_parser("remove", aliases=["rm"], help=help)
    subparser.add_argument("dataset", type=pathlib.Path)
    subparser.set_defaults(func=cmd_remove)

    # url
    help = "URL from where a dataset can be downloaded."
    subparser = subparsers.add_parser("url", help=help)
    subparser.add_argument("--json", action="store_true")
    subparser.add_argument("dataset", type=str)
    subparser.set_defaults(func=cmd_url)

    # info
    help = "Get metadata about a dataset."
    subparser = subparsers.add_parser("info", help=help)
    subparser.add_argument("--json", action="store_true")
    subparser.add_argument("dataset", type=str)
    subparser.set_defaults(func=cmd_info)

    # show
    help = "Display a dataset."
    subparser = subparsers.add_parser("show", help=help)
    subparser.add_argument("--json", action="store_true")
    subparser.add_argument("dataset", type=dataset_with_slice)
    subparser.set_defaults(func=cmd_show)

    # download
    help = "Download a dataset and save it in the local system."
    subparser = subparsers.add_parser("download", help=help)
    subparser.add_argument("--json", action="store_true")
    subparser.add_argument("dataset", type=pathlib.Path)
    subparser.add_argument("localpath", nargs="?", default=None, type=pathlib.Path)
    subparser.set_defaults(func=cmd_download)

    # upload
    help = "Upload a local dataset to subscriber."
    subparser = subparsers.add_parser("upload", help=help)
    subparser.add_argument("localpath", type=pathlib.Path)
    subparser.add_argument("dataset", type=pathlib.Path)
    subparser.set_defaults(func=cmd_upload)

    # adduser
    help = "Add a new user."
    subparser = subparsers.add_parser("adduser", help=help)
    subparser.add_argument("--superuser", "-S", action="store_true", default=False)
    subparser.add_argument("newuser", type=str)
    subparser.add_argument("newpass", nargs="?")
    subparser.set_defaults(func=cmd_adduser)

    # deluser
    help = "Delete a user."
    subparser = subparsers.add_parser("deluser", help=help)
    subparser.add_argument("user", type=str)
    subparser.set_defaults(func=cmd_deluser)

    # listusers
    help = "List all users."
    subparser = subparsers.add_parser("listusers", aliases=["lsu"], help=help)
    subparser.add_argument("--json", action="store_true")
    subparser.add_argument("user", nargs="?")
    subparser.set_defaults(func=cmd_listusers)

    # Go
    args = utils.run_parser(parser)
    client = cat2.Client(args.urlbase, (args.username, args.password))
    args.func(client, args)


if __name__ == "__main__":
    main()



import tempfile
from pathlib import Path

import pytest

try:
    from caterva2.services import hdf5root
except ImportError:
    hdf5root = None


def get_source_dir():
    return Path(__file__).parent.parent.parent


def get_examples_dir():
    return get_source_dir() / 'root-example'


@pytest.fixture(scope='session')
def examples_dir():
    """Directory of read-only example files in Caterva2 source"""
    return get_examples_dir()


def make_examples_hdf5(mkdtemp=lambda: Path(tempfile.mkdtemp())):
    if hdf5root is None:
        return None
    h5fpath = mkdtemp() / 'root-example.h5'
    hdf5root.create_example_root(h5fpath)
    return h5fpath


@pytest.fixture(scope='session')
def examples_hdf5(tmp_path_factory):
    """HDF5 file with example datasets in a new temporary directory"""
    return make_examples_hdf5(lambda: tmp_path_factory.mktemp('hdf5'))



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

import numpy as np
import pytest

from .services import TEST_HDF5_ROOT

hdf5root = pytest.importorskip("caterva2.services.hdf5root", reason="HDF5 support not present")


@pytest.fixture
def sub_urlbase(services):
    return services.get_urlbase("subscriber")


@pytest.fixture
def api_root(client):
    return client.get(TEST_HDF5_ROOT)


def test_not_unsupported(api_root):
    for node in api_root.file_list:
        assert not node.startswith("unsupported/")


def test_ds_name_ext(api_root):
    for node in api_root.file_list:
        node.endswith(".b2nd")  # no other conversions supported yet


def test_scalar(api_root):
    ds = api_root["scalar.b2nd"]
    v = ds[()]
    assert v.dtype.kind == "f"
    assert v == pytest.approx(123.456)


def test_string(api_root):
    ds = api_root["string.b2nd"]
    v = ds[()]
    assert v.dtype.kind == "S"
    assert v == b"Hello world!"


def test_nonchunked(api_root):
    ds = api_root["arrays/2d-nochunks.b2nd"]
    ds_chunks = ds.meta["chunks"]
    assert ds_chunks is not None
    assert len(ds_chunks) == 2
    v = ds[:]
    a = np.arange(100, dtype="complex128").reshape(10, 10)
    a = a + a * 1j
    np.testing.assert_array_equal(v, a)


def test_chunked(api_root):
    ds = api_root["arrays/2d-gzip.b2nd"]
    ds_chunks = tuple(ds.meta["chunks"])
    assert ds_chunks == (4, 4)  # chunk shape is kept
    v = ds[:]
    a = np.arange(100, dtype="complex128").reshape(10, 10)
    a = a + a * 1j
    np.testing.assert_array_equal(v, a)


def test_blosc2(api_root):
    ds = api_root["arrays/3d-blosc2.b2nd"]
    ds_chunks = tuple(ds.meta["chunks"])
    assert ds_chunks == (4, 10, 10)  # chunk shape is kept
    # TODO: compression parameters
    # cparams = ds.meta['schunk']['cparams']
    # assert cparams['codec'] == blosc2.Codec.LZ4.value
    # assert cparams['filters'] == [0, 0, 0, 0, 0,
    #                               blosc2.Filter.BITSHUFFLE.value]
    v = ds[:]
    a = np.arange(1000, dtype="uint8").reshape(10, 10, 10)
    np.testing.assert_array_equal(v, a)


@pytest.mark.parametrize("slice_", [slice(None), 1, slice(2, 6), (slice(None, 6), slice(5, 8), slice(6))])
def test_slicing(api_root, slice_):
    ds = api_root["arrays/3d-blosc2.b2nd"]
    v = ds[:]
    a = np.arange(1000, dtype="uint8").reshape(10, 10, 10)
    np.testing.assert_array_equal(v[slice_], a[slice_])


def test_vlmeta(api_root):
    ds = api_root["attrs.b2nd"]
    assert len(ds.vlmeta) == 9
    m = ds.vlmeta
    assert m["Int"] == m["IntT"] == 42
    # TODO: consistent conversion of strings
    # assert m['Bin'] == m['BinT'] == b'foo'
    # assert m['Str'] == m['StrT'] == 'bar'
    assert m["Arr"] == m["ArrT"] == [[0, 1], [2, 3]]
    # TODO: consistent conversion of strings
    # assert m['NilBin'] == b''
    # assert m['NilStr'] == ''
    assert m["NilInt"] is None



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

import io
import os
import pathlib
from collections.abc import AsyncIterator, Callable, Collection, Iterator

try:
    from typing import Self
except ImportError:  # Python < 3.11
    from typing import TypeVar
    Self = TypeVar('Self', bound='PubRoot')    # noqa: F821

# Requirements
import blosc2
import pydantic
import watchfiles

# Project
from caterva2.services import pubroot, srv_utils


class DirectoryRoot:
    """Represents a publisher root which keeps datasets as files
    in a directory.
    """

    Path = pubroot.PubRoot.Path

    @classmethod
    def get_maker(cls, target: str) -> Callable[[], Self] | None:
        try:
            path = pathlib.Path(target)
            if not path.is_dir():
                return None
        except Exception:
            return None
        return lambda: cls(path)

    def __init__(self, path: pathlib.Path):
        abspath = path.resolve(strict=True)
        # Force an error for non-dirs or non-readable dirs.
        next(abspath.iterdir())

        self.abspath = abspath

    def walk_dsets(self) -> Iterator[Path]:
        return (self.Path(p.relative_to(self.abspath))
                for p in self.abspath.glob('**/*')
                if not p.is_dir())

    def _rel_to_abs(self, relpath: Path) -> pathlib.Path:
        if relpath.is_absolute():
            raise ValueError(f"path is not relative: {str(relpath)!r}")
        # ``.`` is removed on path instantiation, no need to check for it.
        if os.path.pardir in relpath.parts:
            raise ValueError(f"{str(os.path.pardir)!r} not allowed "
                             f"in path: {str(relpath)!r}")
        abspath = self.abspath / relpath
        if not abspath.is_file():
            raise pubroot.NoSuchDatasetError(relpath)
        return abspath

    def exists_dset(self, relpath: Path) -> bool:
        try:
            abspath = self._rel_to_abs(relpath)
        except pubroot.NoSuchDatasetError:
            return False
        return abspath.is_file()

    def get_dset_etag(self, relpath: Path) -> str:
        abspath = self._rel_to_abs(relpath)
        stat = abspath.stat()
        return f'{stat.st_mtime}:{stat.st_size}'

    def get_dset_meta(self, relpath: Path) -> pydantic.BaseModel:
        abspath = self._rel_to_abs(relpath)
        return srv_utils.read_metadata(abspath)

    def get_dset_chunk(self, relpath: Path, nchunk: int) -> bytes:
        abspath = self._rel_to_abs(relpath)
        b2dset = blosc2.open(abspath)
        schunk = getattr(b2dset, 'schunk', b2dset)
        if nchunk > schunk.nchunks:
            raise pubroot.NoSuchChunkError(nchunk)
        return schunk.get_chunk(nchunk)

    def open_dset_raw(self, relpath: Path) -> io.RawIOBase:
        abspath = self._rel_to_abs(relpath)
        return open(abspath, 'rb')

    async def awatch_dsets(self) -> AsyncIterator[Collection[Path]]:
        async for changes in watchfiles.awatch(self.abspath):
            relpaths = {
                self.Path(pathlib.Path(abspath).relative_to(self.abspath))
                for change, abspath in changes}
            yield relpaths


pubroot.register_root_class(DirectoryRoot)


def create_example_root(path):
    """Create an example Caterva2 directory to be used as a root."""
    import pathlib

    import blosc2
    import numpy as np

    # The examples come from ``SPECS.md`` and ``root-example`` content.
    path = pathlib.Path(path)
    path.mkdir(parents=True)

    with open(path / "README.md", "w") as f:
        f.write("# Header example\n"
                "This is a simple example,\n"
                "\n"
                "with several lines,\n"
                "\n"
                "for showing purposes.\n")

    # A SChunk containing a data buffer.
    blosc2.SChunk(chunksize=100, data=b"Hello world!" * 100,
                  urlpath=path / "ds-hello.b2frame", mode="w")

    # A 1D array (int64).
    a = np.arange(1000, dtype="int64")
    blosc2.asarray(a, chunks=(100,), blocks=(10,),
                   urlpath=path / "ds-1d.b2nd", mode="w")

    # A 1D array (6-byte strings).
    a = np.array([b"foobar"] * 1000)
    blosc2.asarray(a, chunks=(100,), blocks=(10,),
                   urlpath=path / "ds-1d-b.b2nd", mode="w")

    # A 1D array (dtype with fields).
    a = np.empty(1000, dtype=[("a", "int32"), ("b", "float64"), ("c", "S10"), ("d", "?")])
    a["a"] = np.arange(1000, dtype="int32")
    a["b"] = np.linspace(0, 1, 1000, dtype="float64")
    a["c"] = np.array([f"foobar{i}" for i in range(1000)], dtype="S10")
    # A field with random booleans
    a["d"] = np.random.default_rng().choice([True, False], 1000)
    blosc2.asarray(a, chunks=(100,), blocks=(10,),
                   urlpath=path / "ds-1d-fields.b2nd", mode="w")

    # A 2D array with 2 fields
    shape = (100, 200)
    npa = np.linspace(0, 1, np.prod(shape), dtype=np.float32).reshape(shape)
    npb = np.linspace(1, 2, np.prod(shape), dtype=np.float64).reshape(shape)
    nps = np.empty(shape, dtype=[("a", npa.dtype), ("b", npb.dtype)])
    nps["a"] = npa
    nps["b"] = npb
    s = blosc2.asarray(nps, mode="w", urlpath=path / "ds-2d-fields.b2nd")
    a = blosc2.NDField(s, "a")
    b = blosc2.NDField(s, "b")

    # A scalar (string) with variable-length metalayers (user attributes).
    a = np.str_("foobar")
    b = blosc2.asarray(a, urlpath=path / "ds-sc-attr.b2nd", mode="w")
    for k, v in {'a': 1, 'b': "foo", 'c': 123.456}.items():
        b.schunk.vlmeta[k] = v

    (path / "dir1").mkdir()

    # A 2D array (uint16).
    a = np.arange(200, dtype="uint16").reshape(10, 20)
    blosc2.asarray(a, chunks=(5, 5), blocks=(2, 3),
                   urlpath=path / "dir1/ds-2d.b2nd", mode="w")

    # A 3D array (float32).
    a = np.arange(60, dtype="float32").reshape(3, 4, 5)
    blosc2.asarray(a, chunks=(2, 3, 4), blocks=(2, 2, 2),
                   urlpath=path / "dir1/ds-3d.b2nd", mode="w")

    (path / "dir2").mkdir()

    # A 4D array (complex128).
    a = np.arange(120, dtype="complex128").reshape(2, 3, 4, 5)
    blosc2.asarray(a + a * 1j, chunks=(1, 2, 3, 4), blocks=(1, 2, 2, 2),
                   urlpath=path / "dir2/ds-4d.b2nd", mode="w")


def main():
    import os
    import sys
    try:
        _, c2dpath = sys.argv
    except ValueError:
        prog = os.path.basename(sys.argv[0])
        print(f"Usage: {prog} CATERVA2_DIR", file=sys.stderr)
        sys.exit(1)
    create_example_root(c2dpath)
    print(f"Created example Caterva2 root: {c2dpath!r}", file=sys.stderr)


if __name__ == '__main__':
    main()



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

# Minimal example on browsing a tree of datasets/files

import pathlib

from textual.app import App, ComposeResult
from textual.widgets import Tree

# Project
from caterva2 import api, api_utils, utils


class TreeApp(App):

    def __init__(self, args):
        super().__init__()
        self.root = args.root
        auth_cookie = None
        if args.username and args.password:
            user_auth = {"username": args.username, "password": args.password}
            auth_cookie = api_utils.get_auth_cookie(args.urlbase, user_auth)
        api.subscribe(args.root, args.urlbase, auth_cookie=auth_cookie)
        self.data = api.get_list(args.root, args.urlbase,
                                 auth_cookie=auth_cookie)

    def compose(self) -> ComposeResult:
        path = self.root / pathlib.Path(self.data[0])
        root, _ = path.parts
        tree: Tree[dict] = Tree(root)
        tree.root.expand()
        datasets = tree.root.add("Datasets", expand=True)
        files = tree.root.add("Files", expand=True)
        for dataset in self.data:
            path = self.root / pathlib.Path(dataset)
            _, *parts = path.parts
            if dataset.endswith((".b2nd", ".b2frame")):
                datasets.add_leaf("/".join(parts))
            else:
                files.add_leaf("/".join(parts))
        yield tree


def main():
    conf = utils.get_conf()
    parser = utils.get_parser()
    parser.add_argument('--subscriber',
                        dest='urlbase', type=utils.urlbase_type,
                        default=conf.get('subscriber.url',
                                         api.sub_urlbase_default))
    parser.add_argument('--username', default=conf.get('client.username'))
    parser.add_argument('--password', default=conf.get('client.password'))
    parser.add_argument('--root', default='foo')

    # Go
    args = utils.run_parser(parser)
    app = TreeApp(args)
    app.run()


if __name__ == "__main__":
    main()



import pytest

from caterva2 import utils


def get_configuration():
    return utils.get_conf()


@pytest.fixture(scope='session')
def configuration():
    """Caterva2 configuration, if available"""
    return get_configuration()



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################


# Requirements
from fastapi import FastAPI
from fastapi.routing import APIRouter
from fastapi_websocket_pubsub import PubSubEndpoint

# Project
from caterva2 import models, utils
from caterva2.services import srv_utils

# State
database = None


# API
app = FastAPI()


@app.get("/api/roots", response_model_exclude_none=True)
async def get_roots() -> dict[str, models.Root]:
    return database.roots


@app.post("/api/roots")
async def post_roots(root: models.Root) -> models.Root:
    database.roots[root.name] = root
    database.save()
    await endpoint.publish(["@new"], root)
    return root


# Pub/Sub interface
router = APIRouter()
endpoint = PubSubEndpoint()
endpoint.register_route(router)
app.include_router(router)


def main():
    conf = utils.get_conf("broker")
    parser = utils.get_parser(
        http=conf.get(".http", "localhost:8000"),
        loglevel=conf.get(".loglevel", "warning"),
        statedir=conf.get(".statedir", "_caterva2/bro"),
    )
    args = utils.run_parser(parser)

    # Init database
    # roots = {name: <Root>}
    statedir = args.statedir.resolve()
    global database
    database = srv_utils.Database(statedir / "db.json", models.Broker(roots={}))
    print(database.data)

    # Run
    srv_utils.uvicorn_run(app, args)


if __name__ == "__main__":
    main()



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

import asyncio
import contextlib
import logging

# Requirements
import blosc2
from fastapi import FastAPI, Response, responses

import caterva2.services.dirroot

# Project
from caterva2 import api_utils, models, utils
from caterva2.services import pubroot, srv_utils

with contextlib.suppress(ImportError):
    import caterva2.services.hdf5root  # noqa: F401


logger = logging.getLogger("pub")

# Configuration
broker = None
name = None
proot = None
nworkers = 1

# State
cache = None
client = None
database = None  # <Database> instance


async def worker(queue):
    while True:
        relpath = await queue.get()
        with utils.log_exception(logger, "Publication failed"):
            assert isinstance(relpath, proot.Path)
            key = str(relpath)
            if proot.exists_dset(relpath):
                print("UPDATE", relpath)
                # Load metadata
                if relpath.suffix in {".b2frame", ".b2nd"}:
                    metadata = proot.get_dset_meta(relpath)
                else:
                    # Compress regular files in publisher's cache
                    with proot.open_dset_raw(relpath) as f:
                        data = f.read()
                    b2path = cache / f"{relpath}.b2"
                    srv_utils.compress(data, b2path)
                    metadata = srv_utils.read_metadata(b2path)

                # Publish
                metadata = metadata.model_dump()
                data = {"path": str(relpath), "metadata": metadata}
                await client.publish(name, data=data)
                # Update database
                database.etags[key] = proot.get_dset_etag(relpath)
                database.save()
            else:
                print("DELETE", relpath)
                data = {"path": str(relpath)}
                await client.publish(name, data=data)
                # Update database
                if key in database.etags:
                    del database.etags[key]
                    database.save()

        queue.task_done()


async def watch_root(queue):
    # On start, notify the network about changes to the datasets, changes done since the
    # last run.
    etags = database.etags.copy()
    for relpath in proot.walk_dsets():
        key = str(relpath)
        val = etags.pop(key, None)
        if val != proot.get_dset_etag(relpath):
            queue.put_nowait(relpath)

    # The etags left are those that were deleted
    for key in etags:
        relpath = proot.Path(key)
        queue.put_nowait(relpath)
        del database.etags[key]
        database.save()

    # Watch root for changes
    async for changes in proot.awatch_dsets():
        for relpath in changes:
            queue.put_nowait(relpath)

    print("THIS SHOULD BE PRINTED ON CTRL+C")


@contextlib.asynccontextmanager
async def lifespan(app: FastAPI):
    # Connect to broker
    global client
    client = srv_utils.start_client(f"ws://{broker}/pubsub")

    # Create queue and start workers
    queue = asyncio.Queue()
    tasks = []
    for _ in range(nworkers):
        task = asyncio.create_task(worker(queue))
        tasks.append(task)

    # Watch dataset files (must wait before publishing)
    await client.wait_until_ready()
    watch_task = asyncio.create_task(watch_root(queue))

    yield

    # Cancel watch task
    watch_task.cancel()

    # Cancel worker tasks
    for task in tasks:
        task.cancel()
    await asyncio.gather(*tasks, return_exceptions=True)

    # Disconnect from broker
    await srv_utils.disconnect_client(client)


app = FastAPI(lifespan=lifespan)


@app.get("/api/list")
async def get_list():
    return list(proot.walk_dsets())


@app.get("/api/info/{path:path}")
async def get_info(
    path: str,
    response: Response,
    if_none_match: srv_utils.HeaderType = None,
):
    relpath = proot.Path(path)
    srv_utils.check_dset_path(proot, relpath)

    # Check etag
    etag = database.etags[str(relpath)]
    if if_none_match == etag:
        return Response(status_code=304)

    if relpath.suffix in {".b2frame", ".b2nd"}:
        meta = proot.get_dset_meta(relpath)
    else:
        b2path = srv_utils.get_abspath(cache, f"{relpath}.b2")
        meta = srv_utils.read_metadata(b2path)

    # Return
    response.headers["Etag"] = etag
    return meta


@app.get("/api/download/{path:path}")
async def get_download(path: str, nchunk: int = -1):
    if nchunk < 0:
        srv_utils.raise_bad_request("Chunk number required")

    relpath = proot.Path(path)
    srv_utils.check_dset_path(proot, relpath)

    if relpath.suffix in {".b2frame", ".b2nd"}:
        chunk = proot.get_dset_chunk(relpath, nchunk)
    else:
        b2path = cache / (f"{relpath}.b2")
        schunk = blosc2.open(b2path)
        chunk = schunk.get_chunk(nchunk)

    downloader = srv_utils.iterchunk(chunk)
    return responses.StreamingResponse(downloader)


def main():
    conf = utils.get_conf("publisher", allow_id=True)
    _stdir = "_caterva2/pub" + (f".{conf.id}" if conf.id else "")
    parser = utils.get_parser(
        broker=conf.get("broker.http", "localhost:8000"),
        http=conf.get(".http", "localhost:8001"),
        loglevel=conf.get(".loglevel", "warning"),
        statedir=conf.get(".statedir", _stdir),
        id=conf.id,
    )
    parser.add_argument("name", nargs="?", default=conf.get(".name"))
    parser.add_argument("root", nargs="?", default=conf.get(".root", "data"))
    args = utils.run_parser(parser)
    if args.name is None:  # because optional positional arg w/o conf default
        raise RuntimeError("root name was not specified in configuration nor in arguments")

    # Global configuration
    global broker, name, proot
    broker = args.broker
    name = args.name
    proot = pubroot.make_root(args.root)

    # Init cache
    global cache
    statedir = args.statedir.resolve()
    cache = statedir / "cache"
    cache.mkdir(exist_ok=True, parents=True)

    # Init database
    global database
    model = models.Publisher(etags={})
    database = srv_utils.Database(statedir / "db.json", model)

    # Register
    data = {"name": name, "http": args.http}
    api_utils.post("/api/roots", json=data, server=args.broker)

    # Run
    srv_utils.uvicorn_run(app, args)


if __name__ == "__main__":
    main()



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

import asyncio
import collections.abc
import contextlib
import functools
import io
import itertools
import json
import logging
import mimetypes
import os
import pathlib
import shutil
import string
import tarfile
import typing
import zipfile

# Requirements
import blosc2
import dotenv
import fastapi
import furl
import httpx
import markdown
import nbconvert
import nbformat
import numpy as np
import PIL.Image

# FastAPI
from fastapi import Depends, FastAPI, Form, Request, UploadFile, responses
from fastapi.responses import FileResponse, HTMLResponse, RedirectResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates

# Project
from caterva2 import api_utils, models, utils
from caterva2.services import settings, srv_utils
from caterva2.services.subscriber import db, schemas, users

BASE_DIR = pathlib.Path(__file__).resolve().parent

# Set CATERVA2_SECRET=XXX in .env file in working directory
dotenv.load_dotenv()

# Logging
logger = logging.getLogger("sub")

# State
clients = {}  # topic: <PubSubClient>
locks = {}

mimetypes.add_type("text/markdown", ".md")  # Because in macOS this is not by default
mimetypes.add_type("application/x-ipynb+json", ".ipynb")


def guess_type(path):
    mimetype, encoding = mimetypes.guess_type(path)
    return mimetype


class PubSourceDataset:
    """
    Class for getting chunks from a dataset on a publisher service.
    """

    def __init__(self, abspath, path, metadata=None):
        self.path = pathlib.Path(path)
        if metadata is not None:
            suffix = abspath.suffix
            if suffix == ".b2nd":
                metadata = models.Metadata(**metadata)
                self._shape = metadata.shape
                self._chunks = metadata.chunks
                self._blocks = metadata.blocks
                dtype = metadata.dtype
                if metadata.dtype.startswith("["):
                    # TODO: eval is dangerous, but we mostly trust the metadata
                    # This is a list, so we need to convert it to a string
                    dtype = eval(dtype)
                self._dtype = np.dtype(dtype)
                self._cparams = dict(metadata.schunk.cparams)
            else:
                if suffix == ".b2frame":
                    metadata = models.SChunk(**metadata)
                else:
                    abspath = pathlib.Path(f"{abspath}.b2")
                    metadata = models.SChunk(**metadata)
                self._typesize = metadata.cparams.typesize
                self._chunksize = metadata.chunksize
                self._nbytes = metadata.nbytes
                self._cparams = dict(metadata.cparams)
            del self._cparams["filters, meta"]
            self._cparams = blosc2.CParams(**self._cparams)
            self.abspath = abspath
            if self.abspath is not None:
                self.abspath.parent.mkdir(exist_ok=True, parents=True)

    def _get_request_args(self, nchunk, return_async_client):
        root, *name = self.path.parts
        root = settings.database.roots[root]
        name = pathlib.Path(*name)

        url = f"/api/download/{name}"
        client, url = api_utils.get_client_and_url(root.http, url, return_async_client=return_async_client)
        args = {"url": url, "params": {"nchunk": nchunk}, "timeout": 5}
        return client, args

    def _get_chunk(self, nchunk):
        client, req_args = self._get_request_args(nchunk, return_async_client=False)

        response = client.get(**req_args)
        response.raise_for_status()
        return response.content

    async def _aget_chunk(self, nchunk):
        client, req_args = self._get_request_args(nchunk, return_async_client=True)

        async with client.stream("GET", **req_args) as resp:
            buffer = []
            async for chunk in resp.aiter_bytes():
                buffer.append(chunk)
            return b"".join(buffer)


# Class representing a NDArray in a publisher
class PubNDDataset(blosc2.ProxyNDSource, PubSourceDataset):
    @property
    def shape(self):
        return self._shape

    @property
    def chunks(self):
        return self._chunks

    @property
    def blocks(self):
        return self._blocks

    @property
    def dtype(self):
        return self._dtype

    @property
    def cparams(self):
        return self._cparams

    def get_chunk(self, nchunk: int) -> bytes:
        return self._get_chunk(nchunk)

    async def aget_chunk(self, nchunk: int) -> bytes:
        return await self._aget_chunk(nchunk)


# Class representing a SChunk in a publisher
class PubSCDataset(blosc2.ProxySource, PubSourceDataset):
    @property
    def typesize(self):
        return self._typesize

    @property
    def chunksize(self):
        return self._chunksize

    @property
    def nbytes(self):
        return self._nbytes

    @property
    def cparams(self):
        return self._cparams

    def get_chunk(self, nchunk: int) -> bytes:
        return self._get_chunk(nchunk)

    async def aget_chunk(self, nchunk: int) -> bytes:
        return await self._aget_chunk(nchunk)


# Factory function for creating a proxy for a dataset in publisher
def PubDataset(abspath, path, metadata=None):
    dataset = PubSourceDataset(abspath, path, metadata)
    # By using __new__() and updating the internal dict of the instance,
    # we can return the right class avoiding calling PubSourceDataset.__init__ again
    if hasattr(dataset, "_shape"):
        # return PubNDDataset(abspath, path, metadata)
        instance = PubNDDataset.__new__(PubNDDataset)
    else:
        # return PubSCDataset(abspath, path, metadata)
        instance = PubSCDataset.__new__(PubSCDataset)
    instance.__dict__.update(dataset.__dict__)
    return instance


def get_disk_usage():
    exclude = {"db.json", "db.sqlite"}
    return sum(path.stat().st_size for path, _ in utils.walk_files(settings.statedir, exclude=exclude))


def truncate_path(path, size=35):
    """
    Smart truncaion of a long path for display.
    """
    assert type(path) is str

    if len(path) < size:
        return path

    # If the path is short just truncate at the end
    parts = pathlib.Path(path).parts
    if len(parts) < 3:
        n = len(path) - size
        return path[:-n] + "..."

    # If the path is long be smarter
    first, last = parts[0], parts[-1]
    label = f"{first}/.../{last}"
    n = len(label) - size
    if n > 0:
        last = last[:-n] + "..."

    return f"{first}/.../{last}"


def make_url(request, name, query=None, **path_params):
    url = request.app.url_path_for(name, **path_params)
    url = str(url)  # <starlette.datastructures.URLPath>
    if query:
        url = furl.furl(url).set(query).url
    return settings.urlbase + url


async def new_root(data, topic):
    logger.info(f"NEW root {topic} {data=}")
    root = models.Root(**data)
    settings.database.roots[root.name] = root
    settings.database.save()


async def updated_dataset(data, topic):
    name = topic
    relpath = data["path"]

    rootdir = settings.cache / name
    abspath = rootdir / relpath
    metadata = data.get("metadata")
    if metadata is None:
        if abspath.suffix not in {".b2nd", ".b2frame"}:
            abspath = pathlib.Path(f"{abspath}.b2")
        if abspath.is_file():
            abspath.unlink()
    else:
        key = f"{name}/{relpath}"
        init_b2(abspath, key, metadata)


def init_b2(abspath, path, metadata):
    dataset = PubDataset(abspath, path, metadata)
    # TODO: not sure if this would prevent some kind of update in dataset. @jdavid?
    # if os.path.exists(dataset.abspath):
    #     return
    schunk_meta = metadata.get("schunk", metadata)
    vlmeta = {}
    for k, v in schunk_meta["vlmeta"].items():
        vlmeta[k] = v
    blosc2.Proxy(dataset, urlpath=dataset.abspath, vlmeta=vlmeta, caterva2_env=True)


def open_b2(abspath, path):
    """
    Open a Blosc2 dataset.

    Return a Proxy if the dataset is in a publisher,
    or the LazyExpr or Blosc2 container otherwise.
    """
    if pathlib.Path(path).parts[0] in {"@personal", "@shared", "@public"}:
        container = blosc2.open(abspath)
        if isinstance(container, blosc2.LazyExpr):
            # Open the operands properly
            operands = container.operands
            for key, value in operands.items():
                if value is None:
                    raise ValueError(f'Missing operand "{key}"')
                if "proxy-source" in value.schunk.meta:
                    # Save operand as Proxy, see blosc2.open doc for more info
                    relpath = srv_utils.get_relpath(
                        value, settings.cache, settings.personal, settings.shared, settings.public
                    )
                    operands[key] = open_b2(value.schunk.urlpath, relpath)
            return container
        else:
            return container

    # Return Proxy
    dataset = PubDataset(abspath, path)
    container = blosc2.open(abspath)
    # No need to pass caterva2_env=True since _cache has already been created
    return blosc2.Proxy(dataset, _cache=container)


#
# Internal API
#


def follow(name: str):
    root = settings.database.roots.get(name)
    if root is None:
        return {name: "This dataset does not exist in the network"}

    if not root.subscribed:
        root.subscribed = True
        settings.database.save()

    # Create root directory in the cache
    rootdir = settings.cache / name
    if not rootdir.exists():
        rootdir.mkdir(exist_ok=True)

    # Get list of datasets
    try:
        data = api_utils.get("/api/list", server=root.http)
    except httpx.ConnectError:
        return None

    # Initialize the datasets in the cache
    for relpath in data:
        # If-None-Match header
        key = f"{name}/{relpath}"
        val = settings.database.etags.get(key)
        headers = None if val is None else {"If-None-Match": val}

        # Call API
        response = api_utils.get(
            f"/api/info/{relpath}",
            headers=headers,
            server=root.http,
            raise_for_status=False,
            return_response=True,
        )
        if response.status_code == 304:
            continue

        response.raise_for_status()
        metadata = response.json()

        # Save metadata and create Proxy
        abspath = rootdir / relpath
        init_b2(abspath, key, metadata)

        # Save etag
        settings.database.etags[key] = response.headers["etag"]
        settings.database.save()

    # Subscribe to changes in the dataset
    if name not in clients:
        client = srv_utils.start_client(f"ws://{settings.broker}/pubsub")
        client.subscribe(name, updated_dataset)
        clients[name] = client

    return None


#
# HTTP API
#


def user_login_enabled():
    if settings.login:
        if not bool(os.environ.get("CATERVA2_SECRET")):
            raise RuntimeError("CATERVA2_SECRET envvar is required")
        return True

    return False


def user_register_enabled():
    if settings.register:
        if not settings.login:
            raise RuntimeError("login config must be enabled")
        return True

    return False


current_active_user = users.current_active_user if user_login_enabled() else (lambda: None)
"""Depend on this if the route needs an authenticated user (if enabled)."""

optional_user = (
    users.fastapi_users.current_user(optional=True, verified=False)  # TODO: set when verification works
    if user_login_enabled()
    else (lambda: None)
)
"""Depend on this if the route may do something with no authentication."""


def _setup_plugin_globals():
    from . import plugins

    # These need to be available for plugins at import time.
    plugins.current_active_user = current_active_user


_setup_plugin_globals()


@contextlib.asynccontextmanager
async def lifespan(app: FastAPI):
    # Initialize the (users) database
    if user_login_enabled():
        await db.create_db_and_tables(settings.statedir)

    # Initialize roots from the broker
    client = None
    if settings.broker:
        try:
            data = api_utils.get("/api/roots", server=settings.broker)
        except httpx.ConnectError:
            logger.warning(f'Broker "{settings.broker}" not available')
        else:
            changed = False
            # Deleted
            d = list(settings.database.roots.items())
            for name, _root in d:
                if name not in data:
                    del settings.database.roots[name]
                    changed = True

            # New or updated
            for name, values in data.items():
                root = models.Root(**values)
                if name not in settings.database.roots:
                    settings.database.roots[root.name] = root
                    changed = True
                elif settings.database.roots[root.name].http != root.http:
                    settings.database.roots[root.name].http = root.http
                    changed = True

            if changed:
                settings.database.save()

            # Follow the @new channel to know when a new root is added
            client = srv_utils.start_client(f"ws://{settings.broker}/pubsub")
            client.subscribe("@new", new_root)

            # Resume following
            for path in settings.cache.iterdir():
                if path.is_dir():
                    follow(path.name)

    yield

    # Disconnect from worker
    if client is not None:
        await srv_utils.disconnect_client(client)


# Visualize the size of a file on a compact and human-readable format
def custom_filesizeformat(value):
    for unit in ["B", "KB", "MB", "GB", "TB"]:
        if value < 1024.0:
            if unit == "B":
                return f"{value:.0f} {unit}"
            return f"{value:.1f} {unit}"
        value /= 1024.0
    return f"{value:.1f} PB"


app = FastAPI(lifespan=lifespan)

# TODO: Support user verification
if user_login_enabled():
    app.include_router(
        users.fastapi_users.get_auth_router(users.auth_backend), prefix="/auth/jwt", tags=["auth"]
    )
    app.include_router(
        users.fastapi_users.get_reset_password_router(),
        prefix="/auth",
        tags=["auth"],
    )


if user_register_enabled():
    app.include_router(
        users.fastapi_users.get_register_router(schemas.UserRead, schemas.UserCreate),
        prefix="/auth",
        tags=["auth"],
    )


def url(path: str) -> str:
    return f"{settings.urlbase}/{path}"


def brand_logo():
    path = "media/logo.webp"
    if not (settings.statedir / path).exists():
        path = "static/logo-caterva2-horizontal-small.webp"

    return url(path)


templates = Jinja2Templates(directory=BASE_DIR / "templates")
templates.env.filters["filesizeformat"] = custom_filesizeformat
templates.env.globals["url"] = url


# Add CSS/JS to templates namespace
BUILD_DIR = "static/build/"
with (BASE_DIR / BUILD_DIR / "manifest.json").open() as file:
    manifest = json.load(file)
    entry = manifest["src/main.js"]
    templates.env.globals["main_css"] = url(BUILD_DIR + entry["css"][0])
    templates.env.globals["main_js"] = url(BUILD_DIR + entry["file"])


@app.get("/api/roots")
async def get_roots(user: db.User = Depends(optional_user)) -> dict:
    """
    Get a dict of roots, with root names as keys and properties as values.

    Returns
    -------
    dict
        The dict of roots.
    """
    # Here we just return the roots that are known by the broker
    # plus the special roots @personal, @shared and @public
    roots = settings.database.roots.copy()
    root = models.Root(name="@public", http="", subscribed=True)
    roots[root.name] = root
    if user:
        for name in ["@personal", "@shared"]:
            root = models.Root(name=name, http="", subscribed=True)
            roots[root.name] = root

    return roots


def get_root(name):
    root = settings.database.roots.get(name)
    if root is None:
        srv_utils.raise_not_found(f"{name} not known by the broker")

    return root


@app.post("/api/subscribe/{name}")
async def post_subscribe(
    name: str,
    user: db.User = Depends(optional_user),
):
    """
    Subscribe to a root.

    Parameters
    ----------
    name : str
        The name of the root.

    Returns
    -------
    str
        'Ok' if successful.
    """
    if name == "@public":
        pass
    elif name in {"@personal", "@shared"}:
        if not user:
            raise srv_utils.raise_unauthorized(f"Subscribing to {name} requires authentication")
    else:
        get_root(name)
        follow(name)
    return "Ok"


@app.get("/api/list/{path:path}")
async def get_list(
    path: pathlib.Path,
    user: db.User = Depends(optional_user),
):
    """
    List the datasets in a root or directory.

    Parameters
    ----------
    path : Path
        The path to a root or directory.

    Returns
    -------
    list
        The list of datasets, as name strings relative to path.
    """
    # Get the root
    root = path.parts[0]
    if root == "@public":
        rootdir = settings.public
    elif root == "@personal":
        if not user:
            srv_utils.raise_not_found("@personal needs authentication")
        rootdir = settings.personal / str(user.id)
    elif root == "@shared":
        if not user:
            srv_utils.raise_not_found("@shared needs authentication")
        rootdir = settings.shared
    else:
        root = get_root(root)
        rootdir = settings.cache / root.name

    # List the datasets in root or directory
    directory = rootdir / pathlib.Path(*path.parts[1:])
    if directory.is_file():
        name = pathlib.Path(directory.name)
        return [str(name.with_suffix("") if name.suffix == ".b2" else name)]
    # Sort the list of datasets and return
    paths = [
        str(relpath.with_suffix("") if relpath.suffix == ".b2" else relpath)
        for _, relpath in utils.walk_files(directory)
    ]
    return sorted(paths)


@app.get("/api/info/{path:path}")
async def get_info(
    path: pathlib.Path,
    user: db.User = Depends(optional_user),
):
    """
    Get the metadata of a dataset.

    Parameters
    ----------
    path : pathlib.Path
        The path to the dataset.

    Returns
    -------
    dict
        The metadata of the dataset.
    """
    abspath, _ = abspath_and_dataprep(path, user=user)
    return srv_utils.read_metadata(
        abspath, settings.cache, settings.personal, settings.shared, settings.public
    )


async def partial_download(abspath, path, slice_=None):
    """
    Download the necessary chunks of a dataset.

    Parameters
    ----------
    abspath : pathlib.Path
        The absolute path to the dataset.
    path : str
        The path to the dataset.
    slice_ : slice, tuple of slices
        The slice to fetch.

    Returns
    -------
    None
        When finished, the dataset is available in cache.
    """
    lock = locks.setdefault(path, asyncio.Lock())
    async with lock:
        proxy = open_b2(abspath, path)
        await proxy.afetch(slice_)


def abspath_and_dataprep(
    path: pathlib.Path, slice_: (tuple | None) = None, user: (db.User | None) = None, may_not_exist=False
) -> tuple[
    pathlib.Path,
    collections.abc.Callable[[], collections.abc.Awaitable],
]:
    """
    Get absolute path in local storage and data preparation operation.

    After awaiting the preparation operation to complete, data in the
    dataset should be ready for reading, either that covered by the slice if
    given, or the whole data otherwise.
    """
    parts = path.parts
    if parts[0] == "@personal":
        if not user:
            raise fastapi.HTTPException(status_code=404)  # NotFound

        filepath = settings.personal / str(user.id) / pathlib.Path(*parts[1:])
        abspath = srv_utils.cache_lookup(settings.personal, filepath, may_not_exist)

        async def dataprep():
            pass

    elif parts[0] == "@shared":
        if not user:
            raise fastapi.HTTPException(status_code=404)  # NotFound

        filepath = settings.shared / pathlib.Path(*parts[1:])
        abspath = srv_utils.cache_lookup(settings.shared, filepath, may_not_exist)

        async def dataprep():
            pass

    elif parts[0] == "@public":
        filepath = settings.public / pathlib.Path(*parts[1:])
        abspath = srv_utils.cache_lookup(settings.public, filepath, may_not_exist)

        async def dataprep():
            pass

    else:
        filepath = settings.cache / path
        abspath = srv_utils.cache_lookup(settings.cache, filepath, may_not_exist)

        async def dataprep():
            return await partial_download(abspath, path, slice_)

    return (abspath, dataprep)


@app.get("/api/fetch/{path:path}")
async def fetch_data(
    path: pathlib.Path,
    slice_: str | None = None,
    user: db.User = Depends(optional_user),
):
    """
    Fetch a dataset.

    Parameters
    ----------
    path : pathlib.Path
        The path to the dataset.
    slice_ : str
        The slice to fetch.

    Returns
    -------
    FileResponse or StreamingResponse
        The (slice of) dataset as a Blosc2 schunk.  When the whole dataset is
        to be downloaded (instead of some slice which does not cover it fully),
        its stored image is served containing all data and metadata (including
        variable length fields).
    """

    slice_ = api_utils.parse_slice(slice_)
    # Download and update the necessary chunks of the schunk in cache
    abspath, dataprep = abspath_and_dataprep(path, slice_, user=user)
    # This is still needed and will only update the necessary chunks
    await dataprep()
    container = open_b2(abspath, path)

    if isinstance(container, blosc2.Proxy):
        container = container._cache

    if isinstance(container, blosc2.NDArray | blosc2.LazyExpr):
        array = container
        schunk = getattr(array, "schunk", None)
        typesize = array.dtype.itemsize
        shape = array.shape
    else:
        # SChunk
        array = None
        schunk = container  # blosc2.SChunk
        typesize = schunk.typesize
        shape = (len(schunk),)
        if isinstance(slice_, int):
            # TODO: make SChunk support integer as slice
            slice_ = slice(slice_, slice_ + 1)

    whole = slice_ is None or slice_ == ()
    if not whole and isinstance(slice_, tuple):
        whole = all(
            isinstance(sl, slice)
            and (sl.start or 0) == 0
            and (sl.stop is None or sl.stop >= sh)
            and sl.step in (None, 1)
            for sl, sh in zip(slice_, shape, strict=False)
        )

    if whole and not isinstance(array, blosc2.LazyExpr):
        # Send the data in the file straight to the client,
        # avoiding slicing and re-compression.
        return FileResponse(abspath, filename=abspath.name, media_type="application/octet-stream")

    if isinstance(array, blosc2.LazyExpr):
        # LazyExpr
        data = array[slice_ or ()]
        data = blosc2.asarray(data)
        data = data.to_cframe()
    elif isinstance(array, blosc2.NDArray):
        # Using NDArray.slice() allows a fast path when it is aligned with the chunks
        # As we are going to serialize the slice right away, it is not clear in which
        # situations a contiguous slice is faster than a non-contiguous one.
        # Let's just use the contiguous one for now, until more testing is done.
        data = array.slice(slice_, contiguous=True).to_cframe()
    else:
        # SChunk
        data = schunk[slice_]  # SChunck => bytes
        # A bytes object can still be compressed as a SChunk
        schunk = blosc2.SChunk(data=data, cparams={"typesize": typesize})
        data = schunk.to_cframe()

    downloader = srv_utils.iterchunk(data)
    return responses.StreamingResponse(downloader, media_type="application/octet-stream")


@app.get("/api/download/{path:path}")
async def download_data(
    path: pathlib.Path,
    # Query parameters
    user: db.User = Depends(optional_user),
):
    async def downloader():
        yield await get_file_content(path, user)

    mimetype = guess_type(path)
    headers = {"Content-Disposition": f'attachment; filename="{path.name}"'}
    return responses.StreamingResponse(downloader(), media_type=mimetype, headers=headers)


html_exporter = nbconvert.HTMLExporter()


@app.get("/api/preview/{path:path}")
async def preview(
    path: pathlib.Path,
    # Query parameters
    width: int | None = None,
    user: db.User = Depends(optional_user),
):
    mimetype = guess_type(path)
    if mimetype.startswith("image/") and width:
        img = await get_image(path, user)

        def downloader():
            yield from resize_image(img, width)

    elif mimetype == "application/x-ipynb+json":
        content = await get_file_content(path, user)
        nb = nbformat.reads(content, as_version=4)
        html, resources = html_exporter.from_notebook_node(nb)
        return HTMLResponse(html)

    else:

        async def downloader():
            yield await get_file_content(path, user)

    return responses.StreamingResponse(downloader(), media_type=mimetype)


@app.get("/api/chunk/{path:path}")
async def get_chunk(
    path: pathlib.PosixPath,
    nchunk: int,
    user: db.User = Depends(optional_user),
):
    abspath, _ = abspath_and_dataprep(path, user=user)
    lock = locks.setdefault(path, asyncio.Lock())
    async with lock:
        if user and path.parts[0] == "@personal":
            container = open_b2(abspath, path)
            if isinstance(container, blosc2.LazyArray):
                # We do not support LazyUDF in Caterva2 yet.
                # In case we do, this would have to be changed.
                chunk = container.get_chunk(nchunk)
            else:
                schunk = getattr(container, "schunk", container)
                chunk = schunk.get_chunk(nchunk)
        else:
            sub_dset = PubDataset(abspath, path)
            chunk = await sub_dset.aget_chunk(nchunk)

    downloader = srv_utils.iterchunk(chunk)
    return responses.StreamingResponse(downloader)


def make_expr(name: str, expr: str, operands: dict[str, str], user: db.User, compute: bool = False) -> str:
    """
    Create a lazy expression dataset in personal space.

    This may raise exceptions if there are problems parsing the dataset name
    or expression, or if the expression refers to operands which have not been
    defined.

    Parameters
    ----------
    name : str
        The name of the dataset to be created (without extension).
    expr : str
        The expression to be evaluated.  It must result in a lazy expression.
    operands : dictionary of strings mapping to strings
        The variables used in the expression and which dataset paths they
        refer to.

    Returns
    -------
    str
        The path of the newly created (or overwritten) dataset.
    """

    if not user:
        raise srv_utils.raise_unauthorized("Creating lazy expressions requires authentication")

    # Parse expression
    name = name.strip()
    expr = expr.strip()
    if not name or not expr:
        raise ValueError("Name or expression should not be empty")
    vars = blosc2.get_expr_operands(expr)

    # Open expression datasets
    var_dict = {}
    for var in vars:
        path = operands[var]
        # Detect special roots
        path = pathlib.Path(path)
        if path.parts[0] == "@personal":
            abspath = settings.personal / str(user.id) / pathlib.Path(*path.parts[1:])
        elif path.parts[0] == "@shared":
            abspath = settings.shared / pathlib.Path(*path.parts[1:])
        elif path.parts[0] == "@public":
            abspath = settings.public / pathlib.Path(*path.parts[1:])
        else:
            abspath = settings.cache / path
        var_dict[var] = open_b2(abspath, path)

    # Create the lazy expression dataset
    arr = blosc2.lazyexpr(expr, var_dict)
    if not isinstance(arr, blosc2.LazyExpr):
        cname = type(arr).__name__
        raise TypeError(f"Evaluates to {cname} instead of lazy expression")

    # Save to filesystem
    path = settings.personal / str(user.id)
    path.mkdir(exist_ok=True, parents=True)
    urlpath = f"{path / name}.b2nd"
    if compute:
        arr.compute(urlpath=urlpath, mode="w")
    else:
        arr.save(urlpath=urlpath, mode="w")

    return f"@personal/{name}.b2nd"


@app.post("/api/lazyexpr/")
async def lazyexpr(
    expr: models.NewLazyExpr,
    user: db.User = Depends(current_active_user),
) -> str:
    """
    Create a lazy expression dataset in personal space.

    The JSON request body must contain a "name" for the dataset to be created
    (without extension), an "expression" to be evaluated, which must result in
    a lazy expression, and an "operands" object which maps variable names used
    in the expression to the dataset paths that they refer to.

    Returns
    -------
    str
        The path of the newly created (or overwritten) dataset.
    """

    def error(msg):
        return fastapi.HTTPException(status_code=400, detail=msg)  # bad request

    try:
        result_path = make_expr(expr.name, expr.expression, expr.operands, user, expr.compute)
    except (SyntaxError, ValueError, TypeError) as exc:
        raise error(f"Invalid name or expression: {exc}") from exc
    except KeyError as ke:
        raise error(f"Expression error: {ke.args[0]} is not in the list of available datasets") from ke
    except RuntimeError as exc:
        raise error(f"Runtime error: {exc}") from exc

    return result_path


@app.post("/api/move/")
async def move(
    payload: models.MoveCopyPayload,
    user: db.User = Depends(current_active_user),
):
    """
    Move a dataset.

    Returns
    -------
    str
        The new path of the dataset.
    """
    if not user:
        raise srv_utils.raise_unauthorized("Moving files requires authentication")

    # Both src and dst should start with a special root
    if not payload.src.startswith(("@personal", "@shared", "@public")):
        raise fastapi.HTTPException(
            status_code=400, detail="Only moving from @personal or @shared or @public roots is allowed"
        )
    if not payload.dst.startswith(("@personal", "@shared", "@public")):
        raise fastapi.HTTPException(
            status_code=400, detail="Only moving to @personal or @shared or @public roots is allowed"
        )
    namepath = pathlib.Path(payload.src)
    destpath = pathlib.Path(payload.dst)
    abspath, _ = abspath_and_dataprep(namepath, user=user)
    dest_abspath, _ = abspath_and_dataprep(destpath, user=user, may_not_exist=True)

    # If destination has not an extension, assume it is a directory
    # If user wants something without an extension, she can add a '.b2' extension :-)
    if dest_abspath.is_dir() or not dest_abspath.suffix:
        dest_abspath /= abspath.name
        destpath /= namepath.name

    if abspath.suffix == ".b2" and dest_abspath.suffix != ".b2":
        dest_abspath = pathlib.Path(f"{dest_abspath}.b2")

    # Not sure if we should allow overwriting, but let's allow it for now
    # if dest_abspath.exists():
    #     raise fastapi.HTTPException(status_code=409, detail="The new path already exists")

    # Make sure the destination directory exists
    dest_abspath.parent.mkdir(exist_ok=True, parents=True)
    abspath.rename(dest_abspath)

    return str(destpath)


@app.post("/api/copy/")
async def copy(
    payload: models.MoveCopyPayload,
    user: db.User = Depends(current_active_user),
):
    """
    Copy a dataset.

    Returns
    -------
    str
        The path of the copied dataset.
    """
    if not user:
        raise srv_utils.raise_unauthorized("Copying files requires authentication")

    src, dst = payload.src, payload.dst
    # src should start with a special root or known root
    if not src.startswith(("@personal", "@shared", "@public")) and src not in settings.database.roots:
        raise fastapi.HTTPException(status_code=400, detail="Only copying from existing roots is allowed")
    # dst should start with a special root
    if not dst.startswith(("@personal", "@shared", "@public")):
        raise fastapi.HTTPException(
            status_code=400, detail="Only copying to @personal or @shared or @public roots is allowed"
        )

    namepath, destpath = pathlib.Path(src), pathlib.Path(dst)
    abspath, _ = abspath_and_dataprep(namepath, user=user)
    dest_abspath, _ = abspath_and_dataprep(destpath, user=user, may_not_exist=True)

    # If destination has not an extension, assume it is a directory
    # If user wants something without an extension, she should add a '.b2' extension
    if dest_abspath.is_dir() or not dest_abspath.suffix:
        dest_abspath /= abspath.name
        destpath /= namepath.name

    if abspath.suffix == ".b2" and dest_abspath.suffix != ".b2":
        dest_abspath = pathlib.Path(f"{dest_abspath}.b2")

    # Not sure if we should allow overwriting, but let's allow it for now
    # if dest_abspath.exists():
    #     raise fastapi.HTTPException(status_code=409, detail="The new path already exists")

    dest_abspath.parent.mkdir(exist_ok=True, parents=True)
    if abspath.is_dir():
        shutil.copytree(abspath, dest_abspath)
    else:
        shutil.copy(abspath, dest_abspath)

    return str(destpath)


@app.post("/api/upload/{path:path}")
async def upload_file(
    path: pathlib.Path,
    file: UploadFile,
    user: db.User = Depends(current_active_user),
):
    """
    Upload a file to a root.

    Parameters
    ----------
    path : pathlib.Path
        The path to store the uploaded file.
    file : UploadFile
        The file to upload.

    Returns
    -------
    str
        The path of the uploaded file.
    """
    if not user:
        raise srv_utils.raise_unauthorized("Uploading requires authentication")

    # Replace the root with absolute path
    root = path.parts[0]
    if root == "@personal":
        abspath = settings.personal / str(user.id) / pathlib.Path(*path.parts[1:])
    elif root == "@shared":
        abspath = settings.shared / pathlib.Path(*path.parts[1:])
    elif root == "@public":
        abspath = settings.public / pathlib.Path(*path.parts[1:])
    else:
        raise fastapi.HTTPException(
            status_code=400,  # bad request
            detail="Only uploading to @personal or @shared or @public roots is allowed",
        )

    # We may upload a new file, or replace an existing file
    if abspath.is_dir():
        abspath /= file.filename
        path /= file.filename

    # Check quota
    # TODO To be fair we should check quota later (after compression, zip unpacking etc.)
    data = await file.read()
    if abspath.suffix not in {".b2", ".b2frame", ".b2nd"}:
        schunk = blosc2.SChunk(data=data)
        newsize = schunk.nbytes
    else:
        newsize = len(data)

    if settings.quota:
        try:
            oldsize = abspath.stat().st_size
        except FileNotFoundError:
            oldsize = 0

        total_size = get_disk_usage() - oldsize + newsize
        if total_size > settings.quota:
            detail = "Upload failed because quota limit has been exceeded."
            raise fastapi.HTTPException(detail=detail, status_code=400)

    # If regular file, compress it
    abspath.parent.mkdir(exist_ok=True, parents=True)
    if abspath.suffix not in {".b2", ".b2frame", ".b2nd"}:
        data = schunk.to_cframe()
        abspath = abspath.with_suffix(abspath.suffix + ".b2")

    # Write the file
    with open(abspath, "wb") as f:
        f.write(data)

    # Return the urlpath
    return str(path)


@app.post("/api/append/{path:path}")
async def append_file(
    path: pathlib.Path,
    file: UploadFile,
    user: db.User = Depends(current_active_user),
):
    """
    Append to dataset (along the first axis).

    Parameters
    ----------
    path : pathlib.Path
        The path to dataset to append.
    file : UploadFile
        The dataset to append.

    Returns
    -------
    tuple
        The new shape of the dataset.
    """
    if not user:
        raise srv_utils.raise_unauthorized("Uploading requires authentication")

    # Replace the root with absolute path
    root = path.parts[0]
    if root == "@personal":
        abspath = settings.personal / str(user.id) / pathlib.Path(*path.parts[1:])
    elif root == "@shared":
        abspath = settings.shared / pathlib.Path(*path.parts[1:])
    elif root == "@public":
        abspath = settings.public / pathlib.Path(*path.parts[1:])
    else:
        detail = "Only uploading to @personal or @shared or @public roots is allowed"
        raise fastapi.HTTPException(detail=detail, status_code=400)

    # We may upload a new file, or replace an existing file
    if not abspath.is_file():
        detail = "Target file does not exist or is not a file"
        raise fastapi.HTTPException(detail=detail, status_code=400)

    if abspath.suffix not in {".b2nd"}:
        detail = "Target file must be a NDArray"
        raise fastapi.HTTPException(detail=detail, status_code=400)

    # Check quota
    # TODO To be fair we should check quota later (after compression, zip unpacking etc.)
    data = await file.read()
    newsize = len(data)

    if settings.quota:
        oldsize = abspath.stat().st_size

        total_size = get_disk_usage() + oldsize + newsize
        if total_size > settings.quota:
            detail = "Upload failed because quota limit has been exceeded."
            raise fastapi.HTTPException(detail=detail, status_code=400)

    # Append the data
    # The original dataset
    orig = blosc2.open(abspath)
    # The data to append is a cframe
    new = blosc2.ndarray_from_cframe(data)
    # Check that the shapes are compatible
    if orig.shape[1:] != new.shape[1:]:
        detail = "The shapes of the original dataset and the data to append are not compatible"
        raise fastapi.HTTPException(detail=detail, status_code=400)
    # Compute the new shape and resize the original dataset
    result_shape = (orig.shape[0] + new.shape[0],) + orig.shape[1:]
    orig.resize(result_shape)
    # Append the new data to orig along the first axis
    orig[orig.shape[0] - new.shape[0] :] = new

    # Return the new shape
    return result_shape


@app.post("/api/remove/{path:path}")
async def remove(
    path: pathlib.Path,
    user: db.User = Depends(current_active_user),
):
    """
    Remove a dataset or a directory path.

    Parameters
    ----------
    path : pathlib.Path
        The path of dataset / directory to remove.

    Returns
    -------
    list
        A list with the paths that have been removed.
    """

    if not user:
        raise srv_utils.raise_unauthorized("Removing files requires authentication")

    # Replace the root with absolute path
    root = path.parts[0]
    if root == "@personal":
        path2 = settings.personal / str(user.id) / pathlib.Path(*path.parts[1:])
    elif root == "@shared":
        path2 = settings.shared / pathlib.Path(*path.parts[1:])
    elif root == "@public":
        path2 = settings.public / pathlib.Path(*path.parts[1:])
    else:
        detail = "Only removing from @personal or @shared or @public roots is allowed"
        raise fastapi.HTTPException(status_code=400, detail=detail)

    # If path2 is a directory, remove the contents of the directory
    if path2.is_dir():
        shutil.rmtree(path2)
    else:
        # Try to unlink the file
        try:
            path2.unlink()
        except FileNotFoundError:
            # Try adding a .b2 extension
            path2 = path2.with_suffix(path2.suffix + ".b2")
            try:
                path2.unlink()
            except FileNotFoundError as exc:
                raise fastapi.HTTPException(
                    status_code=404,  # not found
                    detail="The specified path does not exist",
                ) from exc

    # Return the path
    return path


@app.post("/api/addnotebook/{path:path}")
async def add_notebook(
    path: pathlib.Path,
    user: db.User = Depends(current_active_user),
):
    """
    Add a new notebook.

    Parameters
    ----------
    path : pathlib.Path
        The path where the notebook will be created.

    Returns
    -------
    str
        The path of the new notebook.
    """

    if not user:
        raise srv_utils.raise_unauthorized("Authentication is required")

    if path.suffix != ".ipynb":
        detail = "Notebooks must end with the .ipynb extension"
        raise fastapi.HTTPException(status_code=400, detail=detail)

    # Replace the root with absolute path
    root = path.parts[0]
    if root == "@personal":
        path2 = settings.personal / str(user.id) / pathlib.Path(*path.parts[1:])
    elif root == "@shared":
        path2 = settings.shared / pathlib.Path(*path.parts[1:])
    elif root == "@public":
        path2 = settings.public / pathlib.Path(*path.parts[1:])
    else:
        detail = "Only @personal or @shared or @public roots can be modified"
        raise fastapi.HTTPException(status_code=400, detail=detail)

    # Check a file does not exist in the same path
    path2 = pathlib.Path(f"{path2}.b2")
    if path2.exists():
        detail = "File exists at the given path"
        raise fastapi.HTTPException(status_code=400, detail=detail)

    # Create the new notebook at the given path
    nb = nbformat.v4.new_notebook()
    file = io.StringIO()
    nbformat.write(nb, file)
    data = file.getvalue().encode()
    srv_utils.compress(data, dst=path2)

    return path


#
# HTML interface
#

if user_login_enabled():

    @app.get("/login", response_class=HTMLResponse)
    async def html_login(request: Request, user: db.User = Depends(optional_user)):
        if user:
            return RedirectResponse(settings.urlbase, status_code=307)

        context = {
            "user_register_enabled": user_register_enabled(),
        }
        return templates.TemplateResponse(request, "login.html", context)

    @app.get("/logout", response_class=HTMLResponse)
    async def html_logout(request: Request, user: db.User = Depends(optional_user)):
        if user:
            return RedirectResponse(settings.urlbase, status_code=307)

        return templates.TemplateResponse(request, "logout.html")

    @app.get("/forgot-password", response_class=HTMLResponse)
    async def html_forgot_password(request: Request, user: db.User = Depends(optional_user)):
        if user:
            return RedirectResponse(settings.urlbase, status_code=307)

        return templates.TemplateResponse(request, "forgot-password.html")

    @app.get("/forgot-password-ok", response_class=HTMLResponse)
    async def html_forgot_password_ok(request: Request):
        context = {"settings": settings}
        return templates.TemplateResponse(request, "forgot-password-ok.html", context=context)

    @app.get("/reset-password/{token}", response_class=HTMLResponse, name="html-reset-password")
    async def html_reset_password(request: Request, token: str, user: db.User = Depends(optional_user)):
        if user:
            return RedirectResponse(settings.urlbase, status_code=307)

        context = {"token": token}
        return templates.TemplateResponse(request, "reset-password.html", context)

    @app.post("/api/adduser/")
    async def add_user(
        payload: models.AddUserPayload,
        user: db.User = Depends(current_active_user),
    ):
        """
        Add a user.

        Parameters
        ----------
        payload : AddUserPayload
            The payload containing the username, password and whether the user is a superuser.

        Returns
        -------
        str
            A message indicating success.
        """
        if not user:
            raise srv_utils.raise_unauthorized("Adding a user requires authentication")
        if not user.is_superuser:
            srv_utils.raise_unauthorized("Only superusers can add users")

        # Get the number of current users
        users = await srv_utils.alist_users()
        # None or 0 means unlimited users
        if settings.maxusers and len(users) >= settings.maxusers:
            raise srv_utils.raise_bad_request(f"Only a maximum of {settings.maxusers} users are allowed")

        try:
            await srv_utils.aadd_user(
                payload.username,
                payload.password,
                payload.superuser,
                state_dir=settings.statedir,
            )
        except Exception as exc:
            error_message = str(exc) if str(exc) else exc.__class__.__name__
            raise srv_utils.raise_bad_request(
                f"Error in adding {payload.username}: {error_message}"
            ) from exc
        return f"User added: {payload}"

    @app.get("/api/deluser/{username}")
    async def del_user(
        username: str,
        user: db.User = Depends(current_active_user),
    ):
        """
        Delete a user.

        Parameters
        ----------
        username : str
            The username of the user to delete.

        Returns
        -------
        str
            A message indicating success.
        """
        if not user:
            raise srv_utils.raise_unauthorized("Deleting a user requires authentication")
        if not user.is_superuser:
            srv_utils.raise_unauthorized("Only superusers can delete users")

        try:
            users = await srv_utils.alist_users(username)
            await srv_utils.adel_user(username)
        except Exception as exc:
            error_message = str(exc) if str(exc) else exc.__class__.__name__
            raise srv_utils.raise_bad_request(f"Error in deleting {username}: {error_message}") from exc
        # Remove the personal directory of the user
        userid = str(users[0].id)
        print(f"User {username} with id {userid} has been deleted")
        shutil.rmtree(settings.personal / userid, ignore_errors=True)
        return f"User deleted: {username}"

    @app.get("/api/listusers/")
    async def list_users(
        username: str | None = None,
        user: db.User = Depends(current_active_user),
    ):
        """
        List all users or a specific user.

        Parameters
        ----------
        username : str or None
            The username of the user to list (optional).

        Returns
        -------
        list of dict
            A list of all users (as dictionaries).
        """
        if not user:
            raise srv_utils.raise_unauthorized("Listing users requires authentication")
        return await srv_utils.alist_users(username)

    # TODO: Support user verification


if user_register_enabled():

    @app.get("/register", response_class=HTMLResponse)
    async def html_register(request: Request, user: db.User = Depends(optional_user)):
        if user:
            return RedirectResponse(settings.urlbase, status_code=307)

        return templates.TemplateResponse(request, "register.html")


@app.get("/favicon.ico", include_in_schema=False)
async def favicon():
    return FileResponse(BASE_DIR / "static/logo-caterva2-16x16.png")


@app.get("/", response_class=HTMLResponse)
@app.get("/roots/{path:path}")
async def html_home(
    request: Request,
    path: str = "",
    # Query parameters
    roots: list[str] = fastapi.Query([]),
    search: str = "",
    # Dependencies
    user: db.User = Depends(optional_user),
):
    # Disk usage
    size = get_disk_usage()
    context = {
        "user_login_enabled": user_login_enabled(),
        "roots_url": make_url(request, "htmx_root_list", {"roots": roots}),
        "username": user.email if user else None,
        # Disk usage
        "usage_total": custom_filesizeformat(size),
        # Prompt
        "cmd_url": make_url(request, "htmx_command"),
        "commands": commands_list,
    }

    context["config"] = {}

    if settings.quota:
        context["usage_quota"] = custom_filesizeformat(settings.quota)
        context["usage_percent"] = round((size / settings.quota) * 100)

    if roots:
        paths_url = make_url(request, "htmx_path_list", {"roots": roots, "search": search})
        context["paths_url"] = paths_url

    if path:
        context["meta_url"] = make_url(request, "htmx_path_info", path=path)

    return templates.TemplateResponse(request, "home.html", context)


@app.get("/htmx/root-list/")
async def htmx_root_list(
    request: Request,
    # Query
    roots: list[str] = fastapi.Query([]),
    # Depends
    user: db.User = Depends(optional_user),
):
    context = {
        "checked": roots,
        "roots": sorted(settings.database.roots.values(), key=lambda x: x.name),
        "user": user,
    }
    return templates.TemplateResponse(request, "root_list.html", context)


def _get_rootdir(user, root):
    if root == "@personal":
        if user:
            return settings.personal / str(user.id)
    elif root == "@shared":
        if user:
            return settings.shared
    elif root == "@public":
        return settings.public
    else:
        if not get_root(root).subscribed:
            follow(root)
        return settings.cache / root

    return None


@app.get("/htmx/path-list/", response_class=HTMLResponse)
async def htmx_path_list(
    request: Request,
    # Query parameters
    roots: list[str] = fastapi.Query([]),
    search: str = "",
    # Headers
    hx_current_url: srv_utils.HeaderType = None,
    hx_trigger: srv_utils.HeaderType = None,
    # Depends
    user: db.User = Depends(optional_user),
):
    hx_current_url = furl.furl(hx_current_url)

    # Prepare datasets context
    def get_names():
        n = 1
        while True:
            for name in itertools.product(*[string.ascii_lowercase] * n):
                yield "".join(name)
            n += 1

    names = get_names()

    datasets = []
    query = {"roots": roots, "search": search}

    def add_dataset(path, abspath):
        datasets.append(
            {
                "name": "_",
                "path": path,
                "size": abspath.stat().st_size,
                "url": make_url(request, "html_home", path=path, query=query),
                "label": truncate_path(path),
            }
        )

    for root in roots:
        rootdir = _get_rootdir(user, root)
        for abspath, relpath in utils.walk_files(rootdir):
            if relpath.suffix == ".b2":
                relpath = relpath.with_suffix("")
            path = f"{root}/{relpath}"
            if search in path:
                add_dataset(path, abspath)

    # Add current path if not already in the list
    current_path = hx_current_url.path
    segments = current_path.segments
    if segments and segments[0] == "roots":
        path = str(pathlib.Path(*segments[1:]))
        for dataset in datasets:
            if dataset["path"] == path:
                break
        else:
            root = segments[1]
            rootdir = _get_rootdir(user, root)
            relpath = pathlib.Path(*segments[2:])
            abspath = rootdir / relpath
            if abspath.suffix not in {".b2", ".b2nd", ".b2frame"}:
                abspath = pathlib.Path(f"{abspath}.b2")

            add_dataset(path, abspath)

    # Assign names to datasets
    datasets = sorted(datasets, key=lambda x: x["path"])
    for dataset in datasets:
        dataset["name"] = next(names)

    # Render template
    search_url = make_url(request, "htmx_path_list", {"roots": roots})
    context = {
        "datasets": datasets,
        "search_text": search,
        "search_url": search_url,
        "user": user,
    }
    response = templates.TemplateResponse(request, "path_list.html", context)

    # Push URL only when clicked, not on load/reload
    if hx_trigger != "path-list":
        args = {"roots": roots}
        if search:
            args["search"] = search
        push_url = hx_current_url.set(args).url
        response.headers["HX-Push-Url"] = push_url

    return response


@app.get("/htmx/path-info/{path:path}", response_class=HTMLResponse)
async def htmx_path_info(
    request: Request,
    # Path parameters
    path: pathlib.Path,
    # Headers
    hx_current_url: srv_utils.HeaderType = None,
    hx_trigger: srv_utils.HeaderType = None,
    # Depends
    user: db.User = Depends(optional_user),
):
    # Used to deselect
    if len(path.parts) == 0:
        response = HTMLResponse("")
        push_url = make_url(request, "html_home")
        # Keep query
        current_query = furl.furl(hx_current_url).query
        if current_query:
            push_url = f"{push_url}?{current_query.encode()}"

        response.headers["HX-Push-Url"] = push_url
        return response

    # Read metadata
    abspath, _ = abspath_and_dataprep(path, user=user)
    meta = srv_utils.read_metadata(
        abspath, settings.cache, settings.personal, settings.shared, settings.public
    )

    vlmeta = getattr(getattr(meta, "schunk", meta), "vlmeta", {})
    contenttype = vlmeta.get("contenttype") or guess_dset_ctype(path, meta)
    plugin = plugins.get(contenttype)
    if plugin:
        display = {
            "url": url(f"plugins/{plugin.name}/display/{path}"),
            "label": plugin.label,
        }
    else:
        mimetype = guess_type(path)
        if mimetype and (
            mimetype in {"application/pdf", "application/x-ipynb+json", "text/markdown"}
            or mimetype.startswith("image/")
        ):
            display = {
                "url": url(f"display/{path}"),
                "label": "Display",
            }
        else:
            display = None

    context = {
        "path": path,
        "meta": meta,
        "display": display,
        "can_delete": user and path.parts[0] in {"@personal", "@shared", "@public"},
    }

    # XXX
    if hasattr(meta, "shape"):
        view_url = make_url(request, "htmx_path_view", path=path)
        context.update(
            {
                "view_url": view_url,
                "shape": meta.shape,
            }
        )

    response = templates.TemplateResponse(request, "info.html", context=context)

    # Push URL only when clicked, not on load/reload
    if hx_trigger != "meta":
        push_url = make_url(request, "html_home", path=path)
        # Keep query
        current_query = furl.furl(hx_current_url).query
        if current_query:
            push_url = f"{push_url}?{current_query.encode()}"

        response.headers["HX-Push-Url"] = push_url

    return response


# Global dictionary to store objects
@functools.lru_cache(maxsize=16)
def get_filtered_array(abspath, path, filter, sortby):
    arr = open_b2(abspath, path)
    has_ndfields = hasattr(arr, "fields") and arr.fields != {}
    assert has_ndfields
    idx = None
    sortby = sortby.strip() if sortby else None

    # Filter rows only for NDArray with fields
    if filter:
        # Check whether filter is the name of a field
        if filter in arr.fields:
            if arr.dtype.fields[filter][0] == bool:  # noqa: E721
                # If boolean, give the filter a boolean expression
                filter = f"{filter} == True"
            else:
                raise IndexError("Filter should be a boolean expression")
        # Let's create a LazyExpr with the filter
        larr = arr[filter]
        # TODO: do some benchmarking to see if this is worth it
        idx = larr.indices(sortby).compute()
        # TODO: do some benchmarking to see if a numpy array is faster
        # but be aware that this will consume more memory (uncompressed)
        # idx = larr.indices(sortby)[:]
        arr = larr.sort(sortby).compute()
    elif sortby:
        # NDArray with fields; no need for the compute step
        idx = arr.indices(sortby)
        arr = arr.sort(sortby)

    return arr, idx


@app.post("/htmx/path-view/{path:path}", response_class=HTMLResponse)
async def htmx_path_view(
    request: Request,
    # Path parameters
    path: pathlib.Path,
    # Input parameters
    index: typing.Annotated[list[int] | None, Form()] = None,
    sizes: typing.Annotated[list[int] | None, Form()] = None,
    fields: typing.Annotated[list[str] | None, Form()] = None,
    filter: typing.Annotated[str, Form()] = "",
    sortby: typing.Annotated[str, Form()] = "",
    # Depends
    user: db.User = Depends(optional_user),
):
    abspath, _ = abspath_and_dataprep(path, user=user)
    filter = filter.strip()
    if filter or sortby:
        try:
            arr, idx = get_filtered_array(abspath, path, filter, sortby)
        except TypeError as exc:
            return htmx_error(request, f"Error in filter: {exc}")
        except NameError as exc:
            return htmx_error(request, f"Unknown field: {exc}")
        except ValueError as exc:
            return htmx_error(request, f"ValueError: {exc}")
        except SyntaxError as exc:
            return htmx_error(request, f"SyntaxError: {exc}")
        except IndexError as exc:
            return htmx_error(request, f"IndexError: {exc}")
        except AttributeError as exc:
            return htmx_error(
                request,
                f"Invalid filter: {exc}." f" Only expressions can be used as filters, not field names.",
            )
    else:
        try:
            arr = open_b2(abspath, path)
        except ValueError:
            return htmx_error(
                request, "Cannot open array; if it's a lazy expression, maybe an operand is missing."
            )
        idx = None

    # Local variables
    shape = arr.shape
    ndims = len(shape)

    # Set of dimensions that define the window
    # TODO Allow the user to choose the window dimensions
    has_ndfields = hasattr(arr, "fields") and arr.fields != {}
    dims = list(range(ndims))
    if ndims == 0:
        view_dims = {}
    elif ndims == 1 or has_ndfields:
        view_dims = {dims[-1]}
    else:
        view_dims = {dims[-2], dims[-1]}

    # Default values for input params
    index = (0,) * ndims if index is None else tuple(index)
    if sizes is None:
        sizes = [min(dim, 10) if i in view_dims else 1 for i, dim in enumerate(shape)]

    inputs = []
    tags = []
    for i, (start, size, size_max) in enumerate(zip(index, sizes, shape, strict=False)):
        mod = size_max % size
        start_max = size_max - (mod or size)
        inputs.append(
            {
                "start": start,
                "start_max": start_max,
                "size": size,
                "size_max": size_max,
                "with_size": i in view_dims,
            }
        )
        if inputs[-1]["with_size"]:
            stop = min(start + size, size_max)
            if idx is None:
                tags.append(list(range(start, stop)))
            else:
                tags.append(list(idx[start:stop]))

    if has_ndfields:
        cols = list(arr.fields.keys())
        fields = fields or cols[:5]
        idxs = [cols.index(f) for f in fields]
        rows = [fields]

        # Get array view
        if ndims >= 2:
            arr = arr[index[:-1]]
            i, isize = index[-1], sizes[-1]
            arr = arr[i : i + isize]
            arr = arr.tolist()
        elif ndims == 1:
            i, isize = index[0], sizes[0]
            arr = arr[i : i + isize]
            arr = arr.tolist()
        else:
            arr = [arr[()].tolist()]
        rows += [[row[i] for i in idxs] for row in arr]
    else:
        # Get array view
        cols = None
        if ndims >= 2:
            arr = arr[index[:-2]]
            i, isize = index[-2], sizes[-2]
            j, jsize = index[-1], sizes[-1]
            arr = arr[i : i + isize, j : j + jsize]
            rows = [tags[-1]] + list(arr)
        elif ndims == 1:
            i, isize = index[0], sizes[0]
            arr = [arr[i : i + isize]]
            rows = [tags[-1]] + list(arr)
        else:
            arr = [[arr[()]]]
            rows = list(arr)

    # Render
    context = {
        "view_url": make_url(request, "htmx_path_view", path=path),
        "inputs": inputs,
        "rows": rows,
        "cols": cols,
        "fields": fields,
        "filter": filter,
        "sortby": sortby,
        "shape": shape,
        "tags": tags if len(tags) == 0 else tags[0],
    }
    return templates.TemplateResponse(request, "info_view.html", context)


class AddUserCmd:
    """Add a new user."""

    names = ("adduser",)
    expected = "adduser <username>"
    nargs = 2

    @classmethod
    async def call(cls, request, user, argv, operands, hx_current_url):
        payload = models.AddUserPayload(username=argv[1], password=None, superuser=False)
        message = await add_user(payload, user)
        return htmx_message(request, message)


class DelUserCmd:
    """Remove user."""

    names = ("deluser",)
    expected = "deluser <username>"
    nargs = 2

    @classmethod
    async def call(cls, request, user, argv, operands, hx_current_url):
        message = await del_user(argv[1], user)
        return htmx_message(request, message)


class ListUsersCmd:
    """List users."""

    names = ("lsu", "listusers")
    expected = "lsu/listusers"
    nargs = 1

    @classmethod
    async def call(cls, request, user, argv, operands, hx_current_url):
        lusers = await list_users()
        users = [user.email for user in lusers]
        return htmx_message(request, f"Users: {users}")


class CopyCmd:
    """Copy file."""

    names = ("cp", "copy")
    expected = "cp/copy <src> <dst>"
    nargs = 3

    @classmethod
    async def call(cls, request, user, argv, operands, hx_current_url):
        src, dst = operands.get(argv[1], argv[1]), operands.get(argv[2], argv[2])
        payload = models.MoveCopyPayload(src=src, dst=dst)
        result_path = await copy(payload, user)
        # Redirect to display new dataset
        result_path = await display_first(result_path, user)
        url = make_url(request, "html_home", path=result_path)
        return htmx_redirect(hx_current_url, url)


class MoveCmd:
    """Move or rename file."""

    names = ("mv", "move")
    expected = "mv/move <src> <dst>"
    nargs = 3

    @classmethod
    async def call(cls, request, user, argv, operands, hx_current_url):
        src, dst = operands.get(argv[1], argv[1]), operands.get(argv[2], argv[2])
        payload = models.MoveCopyPayload(src=src, dst=dst)
        result_path = await move(payload, user)
        # Redirect to display new dataset
        result_path = await display_first(result_path, user)
        url = make_url(request, "html_home", path=result_path)
        return htmx_redirect(hx_current_url, url)


class RemoveCmd:
    """Remove file."""

    names = ("rm", "remove")
    expected = "rm/remove <path>"
    nargs = 2

    @classmethod
    async def call(cls, request, user, argv, operands, hx_current_url):
        path = operands.get(argv[1], argv[1])
        path = pathlib.Path(path)
        await remove(path, user)
        response = responses.Response(status_code=204)
        response.headers["HX-Refresh"] = "true"
        return response


class AddNotebookCmd:
    """Add a new notebook."""

    names = ("addnb",)
    expected = "addnb <path>"
    nargs = 2

    @classmethod
    async def call(cls, request, user, argv, operands, hx_current_url):
        path = pathlib.Path(argv[1])
        path = await add_notebook(path, user)

        # Redirect to display new dataset
        url = make_url(request, "html_home", path=path)
        return htmx_redirect(hx_current_url, url)


commands_list = [AddUserCmd, DelUserCmd, ListUsersCmd, CopyCmd, MoveCmd, RemoveCmd, AddNotebookCmd]

commands = {}
for cmd in commands_list:
    for name in cmd.names:
        if name in commands:
            raise ValueError(f'duplicated "{name}" command')
        commands[name] = cmd


@app.post("/htmx/command/", response_class=HTMLResponse)
async def htmx_command(
    request: Request,
    # Body
    command: typing.Annotated[str, Form()],
    names: typing.Annotated[list[str] | None, Form()] = None,
    paths: typing.Annotated[list[str] | None, Form()] = None,
    # Headers
    hx_current_url: srv_utils.HeaderType = None,
    # Depends
    user: db.User = Depends(current_active_user),
):
    if names is None:
        names = []

    if paths is None:
        paths = []

    operands = dict(zip(names, paths, strict=False))
    argv = command.split()

    # First check for expressions
    nargs = len(argv)
    if nargs == 0:
        return responses.Response(status_code=204)

    elif nargs > 1 and argv[1] in {"=", ":="}:
        operator = argv[1]
        compute = operator == ":="
        try:
            result_name, expr = command.split(operator, maxsplit=1)
            result_path = make_expr(result_name, expr, operands, user, compute=compute)
            url = make_url(request, "html_home", path=result_path)
            return htmx_redirect(hx_current_url, url)
        except SyntaxError:
            return htmx_error(request, "Invalid syntax: expected <varname> = <expression>")
        except ValueError as exc:
            return htmx_error(request, f"Invalid expression: {exc}")
        except TypeError as exc:
            return htmx_error(request, f"Invalid expression: {exc}")
        except KeyError as exc:
            error = f"Expression error: {exc.args[0]} is not in the list of available datasets"
            return htmx_error(request, error)
        except RuntimeError as exc:
            return htmx_error(request, f"Runtime error: {exc}")

    # Commands
    cmd = commands.get(argv[0])
    if cmd is not None:
        if len(argv) != cmd.nargs:
            return htmx_error(request, f"Invalid syntax: expected {cmd.expected}")

        try:
            return await cmd.call(request, user, argv, operands, hx_current_url)
        except Exception as exc:
            return htmx_error(request, f'Error in "{command}" command: {exc.detail}')

    # If the command is not recognized
    return htmx_error(request, f'Invalid command "{argv[0]}" or expression not found')


async def display_first(result_path, user):
    paths = await get_list(pathlib.Path(result_path), user)
    if len(paths) > 1:
        # Display the first path found
        result_path = f"{result_path}/{paths[0]}"
    elif len(paths) == 1 and not result_path.endswith(paths[0]):
        result_path = f"{result_path}/{paths[0]}"
    return result_path


def htmx_message(request, msg):
    context = {"message": msg}
    return templates.TemplateResponse(request, "message.html", context, status_code=400)


def htmx_error(request, msg, status_code=400):
    context = {"error": msg}
    return templates.TemplateResponse(request, "error.html", context, status_code=status_code)


def htmx_redirect(current_url, target_url, root=None):
    response = responses.JSONResponse("OK")
    query = furl.furl(current_url).query
    roots = query.params.getlist("roots")

    if root and root not in roots:
        query = query.add({"roots": root})

    response.headers["HX-Redirect"] = f"{target_url}?{query.encode()}"
    return response


@app.post("/htmx/upload/{name}")
async def htmx_upload(
    request: Request,
    name: str,
    # Body
    file: UploadFile,
    # Headers
    hx_current_url: srv_utils.HeaderType = None,
    # Depends
    user: db.User = Depends(current_active_user),
):
    if not user:
        raise srv_utils.raise_unauthorized("Uploading files requires authentication")

    if name == "@personal":
        path = settings.personal / str(user.id)
    elif name == "@shared":
        path = settings.shared
    elif name == "@public":
        path = settings.public
    else:
        raise fastapi.HTTPException(status_code=404)  # NotFound

    # Read the file and check quota
    data = await file.read()
    if settings.quota:
        total_size = get_disk_usage() + len(data)
        if total_size > settings.quota:
            error = "Upload failed because quota limit has been exceeded."
            return htmx_error(request, error)

    path.mkdir(exist_ok=True, parents=True)
    filename = pathlib.Path(file.filename)

    # If a tarball or zipfile, extract the files in path
    # We also filter out hidden files and MacOSX metadata
    suffix = filename.suffix
    suffixes = filename.suffixes[-2:]
    if suffix in [".tar", ".tgz", ".zip"] or suffixes == [".tar", ".gz"]:
        file.file.seek(0)  # Reset file pointer
        if suffix == ".zip":
            with zipfile.ZipFile(file.file, "r") as archive:
                members = [
                    m
                    for m in archive.namelist()
                    if (
                        not os.path.basename(m).startswith(".")
                        and not os.path.basename(m).startswith("__MACOSX")
                    )
                ]
                archive.extractall(path, members=members)
                # Convert members elements to Path instances
                members = [pathlib.Path(m) for m in members]
        else:
            mode = "r:gz" if suffix in {".tgz", ".gz"} else "r"
            with tarfile.open(fileobj=file.file, mode=mode) as archive:
                members = [
                    m
                    for m in archive.getmembers()
                    if (
                        not os.path.basename(m.name).startswith(".")
                        and not os.path.basename(m.name).startswith("__MACOSX")
                    )
                ]
                archive.extractall(path, members=members)
                # Convert members elements to Path instances
                members = [pathlib.Path(m.name) for m in members]

        # Compress files that are not compressed yet
        new_members = [
            member
            for member in members
            if not (path / member).is_dir() and member.suffix not in {".b2", ".b2frame", ".b2nd"}
        ]
        for member in new_members:
            srv_utils.compress_file(path / member)

        # We are done, redirect to home, and show the new files, starting with the first one
        first_member = next((m for m in new_members), None)
        path = f"{name}/{first_member}"
        return htmx_redirect(hx_current_url, make_url(request, "html_home", path=path), root=name)

    if filename.suffix not in {".b2", ".b2frame", ".b2nd"}:
        schunk = blosc2.SChunk(data=data)
        data = schunk.to_cframe()
        filename = f"{filename}.b2"

    # Save file
    with open(path / filename, "wb") as dst:
        dst.write(data)

    # Redirect to display new dataset
    path = f"{name}/{filename}"
    if path.endswith(".b2"):
        path = path[:-3]
    url = make_url(request, "html_home", path=path)
    return htmx_redirect(hx_current_url, url, root=name)


@app.delete("/htmx/delete/{path:path}", response_class=HTMLResponse)
async def htmx_delete(
    request: Request,
    # Path parameters
    path: pathlib.Path,
    # Headers
    hx_current_url: srv_utils.HeaderType = None,
    # Depends
    user: db.User = Depends(current_active_user),
):
    # Find absolute path to file
    parts = list(path.parts)
    name = parts[0]
    if name == "@personal":
        parts[0] = str(user.id)
        path = pathlib.Path(*parts)
        abspath = settings.personal / path
    elif name == "@shared":
        path = pathlib.Path(*parts[1:])
        abspath = settings.shared / path
    elif name == "@public":
        path = pathlib.Path(*parts[1:])
        abspath = settings.public / path
    else:
        return fastapi.HTTPException(status_code=400)

    # Remove
    if abspath.suffix not in {".b2frame", ".b2nd"}:
        abspath = abspath.with_suffix(abspath.suffix + ".b2")
        if not abspath.exists():
            return fastapi.HTTPException(status_code=404)
    abspath.unlink()

    # Redirect to home
    url = make_url(request, "html_home")
    return htmx_redirect(hx_current_url, url, root=name)


async def get_container(path, user):
    abspath, dataprep = abspath_and_dataprep(path, user=user)
    await dataprep()
    return open_b2(abspath, path)


async def get_file_content(path, user):
    container = await get_container(path, user)

    # Return compressed (this is used only by the downloader)
    if path.suffix in {".b2frame", ".b2nd"}:
        return container.to_cframe()

    # Return uncompressed
    return container[:]


async def get_image(path, user):
    content = await get_file_content(path, user)
    return PIL.Image.open(io.BytesIO(content))


def resize_image(img, width):
    if width and img.width > width:
        height = (img.height * width) // img.width
        img = img.resize((width, height))

    img_file = io.BytesIO()
    img.save(img_file, format="PNG")
    img_file.seek(0)
    return img_file


@app.get("/display/{path:path}", response_class=HTMLResponse)
async def html_display(
    request: Request,
    # Path parameters
    path: pathlib.Path,
    user: db.User = Depends(optional_user),
    # Response
    response_class=HTMLResponse,
):
    mimetype = guess_type(path)
    if mimetype == "application/pdf":
        data = f"{url('api/preview/')}{path}"
        return f'<object data="{data}" type="application/pdf" class="w-100" style="height: 768px"></object>'
    elif mimetype == "application/x-ipynb+json":
        href = url(f"static/jupyterlite/notebooks/index.html?path={path}")
        src = f"{url('api/preview/')}{path}"
        return (
            f'<a href="{href}" target="_blank" class="btn btn-primary mb-1"><i class="fa-solid fa-gear"></i> Run</a>'
            f'<iframe src="{src}" class="w-100" height="768px"></iframe>'
        )
    elif mimetype == "text/markdown":
        content = await get_file_content(path, user)
        return markdown.markdown(content.decode("utf-8"))
    elif mimetype.startswith("image/"):
        src = f"{url('api/preview/')}{path}"
        img = await get_image(path, user)

        width = 768  # Max size
        links = []
        if img.width > width:
            links.append(
                {
                    "href": src,
                    "label": f"{img.width} x {img.height} (original size)",
                    "target": "blank_",
                }
            )
            src = f"{src}?{width=}"

        context = {"src": src, "links": links}
        return templates.TemplateResponse(request, "display_image.html", context=context)

    return "Format not supported"


#
# For Jupyterlite
#


@app.get("/static/jupyterlite/api/contents/{path:path}")
async def jupyterlite_contents(
    request: Request,
    # Path parameters
    path: pathlib.Path,
    user: db.User = Depends(optional_user),
):
    """
    See https://jupyter-server.readthedocs.io/en/latest/developers/rest-api.html#get--api-contents-path
    """

    # The path must end with all.json
    parts = path.parts
    if parts[-1] != "all.json":
        raise fastapi.HTTPException(status_code=404)  # NotFound

    parts = parts[:-1]
    path = pathlib.Path(*parts)

    # Helper function for directories
    def directory(abspath, relpath, content=None):
        stat = abspath.stat()
        return {
            "content": content,
            "created": utils.epoch_to_iso(stat.st_ctime),
            "format": None if content is None else "json",
            "hash": None,
            "hash_algorithm": None,
            "last_modified": utils.epoch_to_iso(stat.st_mtime),
            "mimetype": None,
            "name": pathlib.Path(relpath).name,
            "path": relpath,
            "size": None,
            "type": "directory",
            "writable": False,
        }

    content = []
    if len(parts) == 0:
        rootdir = _get_rootdir(user, "@personal")
        if rootdir is not None:
            rootdir.mkdir(exist_ok=True)
            content.append(directory(rootdir, "@personal"))

        rootdir = _get_rootdir(user, "@shared")
        if rootdir is not None:
            content.append(directory(rootdir, "@shared"))

        rootdir = _get_rootdir(user, "@public")
        if rootdir is not None:
            content.append(directory(rootdir, "@public"))

        # TODO pub/sub roots: settings.database.roots.values()
        dir_abspath = rootdir.parent
        dir_relpath = ""
    else:
        # Check access to the root
        root, *subpath = parts
        rootdir = _get_rootdir(user, root)
        if rootdir is None:
            raise fastapi.HTTPException(status_code=404)  # NotFound

        # Get absolute and relative paths to the directory
        dir_abspath = rootdir / pathlib.Path(*subpath)
        dir_relpath = path

        for abspath, relpath in utils.iterdir(dir_abspath):
            relpath = path / relpath
            if abspath.is_dir():
                content.append(directory(abspath, relpath))
            elif abspath.is_file():
                if relpath.suffix == ".b2":
                    relpath = relpath.with_suffix("")

                mimetype = guess_type(relpath)
                if mimetype == "application/x-ipynb+json":
                    content_type = "notebook"
                    writable = bool(user)
                else:
                    content_type = "file"
                    writable = False

                stat = abspath.stat()
                content.append(
                    {
                        "content": None,
                        "created": utils.epoch_to_iso(stat.st_ctime),
                        "format": None,
                        "hash": None,
                        "hash_algorithm": None,
                        "last_modified": utils.epoch_to_iso(stat.st_mtime),
                        "mimetype": mimetype,
                        "name": relpath.name,
                        "path": relpath,
                        "size": stat.st_size,  # XXX Return the uncompressed size?
                        "type": content_type,
                        "writable": writable,
                    }
                )
            else:
                raise NotImplementedError("Only directories and files are supported")

    return directory(dir_abspath, dir_relpath, content=content)


@app.get("/static/jupyterlite/files/{path:path}")
async def jupyterlite_files(
    request: Request,
    # Path parameters
    path: pathlib.Path,
    user: db.User = Depends(optional_user),
):
    async def downloader():
        yield await get_file_content(path, user)

    mimetype = guess_type(path)
    return responses.StreamingResponse(downloader(), media_type=mimetype)


@app.get("/service-worker.js")
async def jupyterlite_worker(
    # Query parameters
    enableCache: bool | None = None,
):
    abspath = BASE_DIR / "static/jupyterlite/service-worker.js"
    return FileResponse(abspath, filename=abspath.name, media_type="application/javascript")


@app.get("/api/service-worker-heartbeat", response_class=responses.PlainTextResponse)
async def jupyter_heartbeat():
    return "ok"


#
# Static
#

app.mount("/static", StaticFiles(directory=BASE_DIR / "static"), name="static")


#
# Command line interface
#

plugins = {}


def guess_dset_ctype(path: pathlib.Path, meta) -> str | None:
    """Try to guess dataset's content type (given path and metadata)."""
    for ctype, plugin in plugins.items():
        if hasattr(plugin, "guess") and plugin.guess(path, meta):
            return ctype
    return None


def main():
    # Read configuration file
    conf = utils.get_conf("subscriber", allow_id=True)

    # Parse command line arguments
    _stdir = "_caterva2/sub" + (f".{conf.id}" if conf.id else "")
    parser = utils.get_parser(
        broker=conf.get("broker.http", ""),
        http=conf.get(".http", "localhost:8002"),
        loglevel=conf.get(".loglevel", "warning"),
        statedir=conf.get(".statedir", _stdir),
        id=conf.id,
    )
    args = utils.run_parser(parser)
    settings.broker = args.broker

    # Init cache
    settings.statedir = args.statedir.resolve()
    settings.cache = settings.statedir / "cache"
    settings.cache.mkdir(exist_ok=True, parents=True)
    # Use `download_cached()`, `StaticFiles` does not support authorization.
    # app.mount("/files", StaticFiles(directory=cache), name="files")

    # Shared/Public dirs
    settings.shared = settings.statedir / "shared"
    settings.shared.mkdir(exist_ok=True, parents=True)
    settings.public = settings.statedir / "public"
    settings.public.mkdir(exist_ok=True, parents=True)

    # personal dir
    settings.personal = settings.statedir / "personal"
    settings.personal.mkdir(exist_ok=True, parents=True)
    # Use `download_personal()`, `StaticFiles` does not support authorization.
    # app.mount("/personal", StaticFiles(directory=settings.personal), name="personal")

    # Init database
    model = models.Subscriber(roots={}, etags={})
    settings.database = srv_utils.Database(settings.statedir / "db.json", model)

    # Register display plugins
    from .plugins import tomography  # delay module load

    app.mount(f"/plugins/{tomography.name}", tomography.app)
    plugins[tomography.contenttype] = tomography
    tomography.init(abspath_and_dataprep, settings.urlbase)

    # Mount media
    media = settings.statedir / "media"
    media.mkdir(exist_ok=True, parents=True)
    app.mount("/media", StaticFiles(directory=media), name="media")
    templates.env.globals["brand"] = {
        "logo": brand_logo(),
    }

    # Run
    root_path = str(furl.furl(settings.urlbase).path)
    srv_utils.uvicorn_run(app, args, root_path=root_path)


if __name__ == "__main__":
    main()



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

"""Export a Caterva2 root to datasets in an HDF5 file.

The content of the existing Caterva2 root must be readable, while the HDF5
file is created anew (it must not exist).  For each directory under the
Caterva2 root, a group with the same path name is created in the HDF5 file.
For each file or Blosc2 dataset, an HDF5 dataset with the same path name
(minus any ``.b2*`` extension) is created in the HDF5 file.

Blosc2 arrays (``*.b2nd``), frames (``*.b2frame``) and compressed files
(``*.b2``) have their chunk sizes respected during conversion, while plain
files use a default chunk size.  Arrays result in typed multidimensional
datasets, while the rest result in flat datasets of bytes.  Currently, all
exported datasets use some default Blosc2 compression parameters.

Warning: For the moment, the data in plain files dataset is read and
decompressed into memory in its entirety.

Datasets or attributes which are unsupported or fail to be converted are
simply reported and skipped, and they do not cause the program to fail.
"""

import functools
import logging
import operator
import os
import pathlib
import sys
from collections.abc import Callable, Iterator, Mapping

import blosc2
import h5py
import hdf5plugin
import numpy

from .. import hdf5

# Set to empty mapping to store files as uncompressed HDF5 datasets.
file_h5_compargs = hdf5plugin.Blosc2(cname='zstd', clevel=5, filters=1)


def export_leaf(c2_leaf: os.DirEntry, h5_group: h5py.Group) -> None:
    """Export Caterva2 leaf entry `c2_leaf` into
    open HDF5 group `h5_group`.
    """
    try:
        (h5mkempty, h5_chunks, h5_attrs) = (
            h5mkempty_h5chunkit_h5attrs_from_leaf(c2_leaf))
    except Exception as e:
        logging.error(f"Failed to translate Blosc2 dataset: "
                      f"{c2_leaf.path!r} -> {e!r}")
        return

    try:
        h5_dataset = h5mkempty(h5_group)
        for (chunk_slice, chunk) in zip(h5_dataset.iter_chunks(), h5_chunks, strict=False):
            # Cannot use ``h5_dataset.id.get_chunk_info(nchunk)``
            # as there are indeed no chunks yet.
            chunk_offset = tuple(cs.start for cs in chunk_slice)
            h5_dataset.id.write_direct_chunk(chunk_offset, chunk)
    except Exception as e:
        logging.error(f"Failed to save dataset in HDF5 group "
                      f"{h5_group.name!r}: {c2_leaf.name!r} -> {e!r}")
        return

    for (aname, avalue) in h5_attrs.items():
        try:
            h5_dataset.attrs[aname] = avalue
            logging.info(f"Exported dataset attribute "
                         f"{aname!r}: {h5_dataset.name!r}")
        except Exception as e:
            logging.error(f"Failed to export dataset attribute "
                          f"{aname!r}: {h5_dataset.name!r} -> {e!r}")

    logging.info(f"Exported dataset: {c2_leaf.path!r} => {h5_dataset.name!r}")


def h5compargs_from_b2(b2_array: blosc2.NDArray | blosc2.SChunk) -> Mapping:
    ndim = getattr(b2_array, 'ndim', -1)
    if ndim == 0:  # scalar, no filters/compression allowed
        return {}

    b2_schunk = getattr(b2_array, 'schunk', b2_array)
    cparams = b2_schunk.cparams
    # This is what hdf5plugin does (more or less).
    # Option list as per ``hdf5-blosc2/src/blosc2_filter.c``.
    #
    # Note: These HDF5 filter options (``cd_values``)
    # are just an approximation to the actual cparams/dparams
    # in the Blosc2 super-chunks stored as HDF5 chunks,
    # as they are not yet able to encode all Blosc2 features
    # (like the ordered filter pipeline and meta as of 2024-02).
    # The cparams/dparams in each schunk are more reliable.
    opts = (
        1,  # filter revision
        cparams['blocksize'],  # block size (in bytes)
        cparams['typesize'],  # type size (in bytes)
        b2_schunk.chunksize,  # chunk size (in bytes)
        cparams['clevel'],  # compression level
        # Just a coarse hint that filters are used (see note above),
        # e.g. their order and meta are lost here.
        functools.reduce(operator.or_,  # shuffle method
                         (s.value for s in cparams['filters'])),
        cparams['codec'].value,  # compressor code
    )
    if ndim > 1:
        opts += (
            ndim,  # chunk rank (number of dimensions)
            *b2_array.chunks,  # length of chunk dimension i
        )
    return {'compression': hdf5.BLOSC2_HDF5_FID, 'compression_opts': opts}


def h5mkempty_h5chunkit_h5attrs_from_leaf(c2_leaf: os.DirEntry) -> (
        Callable[[h5py.Group, ...], h5py.Dataset],
        Iterator[bytes],
        Mapping):
    # TODO: mark array / frame / file distinguishably
    h5_args = {}
    h5_attrs = {}

    # TODO: handle symlinks safely
    c2_leaf_name = pathlib.Path(c2_leaf.name)
    if c2_leaf_name.suffix == '.b2nd':
        b2_array = blosc2.open(c2_leaf.path)
        h5_args |= dict(
            name=pathlib.Path(c2_leaf).stem,
            shape=b2_array.shape,
            dtype=b2_array.dtype,
            chunks=(b2_array.chunks if b2_array.ndim > 0 else None),
            **h5compargs_from_b2(b2_array),
        )
        h5_chunkit = h5chunkit_from_array(b2_array)
        h5_attrs |= b2_array.schunk.vlmeta.getall()  # copy avoids SIGSEGV
    elif c2_leaf_name.suffix in ['.b2frame', '.b2']:
        b2_schunk = blosc2.open(c2_leaf.path)
        h5_args |= dict(
            name=pathlib.Path(c2_leaf).stem,
            shape=(len(b2_schunk),),
            dtype=(numpy.uint8 if b2_schunk.typesize == 1
                   else numpy.dtype(f'|V{b2_schunk.typesize}')),
            chunks=(b2_schunk.chunkshape,),
            **h5compargs_from_b2(b2_schunk),
        )
        h5_chunkit = h5chunkit_from_schunk(b2_schunk)
        h5_attrs |= b2_schunk.vlmeta.getall()  # copy avoids SIGSEGV
    else:
        # TODO: do not slurp & re-compress
        with open(c2_leaf, 'rb') as f:
            data = f.read()
        h5_args |= dict(
            name=pathlib.Path(c2_leaf).name,
            shape=(c2_leaf.stat(follow_symlinks=True).st_size,),
            dtype=numpy.uint8,
            data=numpy.frombuffer(data, dtype=numpy.uint8),
            chunks=True,
            **file_h5_compargs,
        )
        h5_chunkit = iter([])  # TODO: iterate over chunks

    def h5_make_empty(h5_group: h5py.Group, **kwds) -> h5py.Dataset:
        dtype = h5_args.pop('dtype')
        return h5_group.create_dataset(
            dtype=dtype,  # not allowed as a real keyword argument
            **(h5_args | kwds)
        )

    return h5_make_empty, h5_chunkit, h5_attrs


def h5chunkit_from_temp(b2_tmp: (blosc2.NDArray | blosc2.SChunk),
                        b2_src: (blosc2.NDArray | blosc2.SChunk)) -> (
        Iterator[bytes]):
    # Blosc2 chunks are passed as they are,
    # but chunks need to be wrapped in a super-chunk
    # to store that as the HDF5 chunk.
    # Copy each Blosc2 chunk into chunk 0 of compatible Blosc2 array,
    # then dump the resulting array/super-chunk.
    # Thus, only one chunk worth of data is kept in memory.
    b2_tmp_schunk = getattr(b2_tmp, 'schunk', b2_tmp)
    b2_src_schunk = getattr(b2_src, 'schunk', b2_src)
    for (b2_nchunk, _) in enumerate(b2_src_schunk.iterchunks_info()):
        b2_tmp_schunk.update_chunk(0, b2_src_schunk.get_chunk(b2_nchunk))
        yield b2_tmp.to_cframe()  # whole array/schunk, not just chunk


def h5chunkit_from_array(b2_array: blosc2.NDArray) -> Iterator[bytes]:
    # See comment in `h5chunkit_from_temp()`.
    tmp_array = blosc2.empty(
        shape=b2_array.chunks,  # note that shape is chunkshape
        dtype=b2_array.dtype,
        chunks=b2_array.chunks,
        blocks=b2_array.blocks,
        cparams=b2_array.schunk.cparams,
        dparams=b2_array.schunk.dparams,
    )
    # Passing `b2_array` instead of `b2_array.schunk`
    # avoids a potential SIGSEGV.
    return h5chunkit_from_temp(tmp_array, b2_array)


def h5chunkit_from_schunk(b2_schunk: blosc2.SChunk) -> Iterator[bytes]:
    # See comment in `h5chunkit_from_temp()`.
    tmp_schunk = blosc2.SChunk(
        chunksize=b2_schunk.chunksize,
        cparams=b2_schunk.cparams,
        dparams=b2_schunk.dparams,
    )
    tmp_schunk.fill_special(b2_schunk.chunkshape,  # note just one chunk
                            blosc2.SpecialValue.UNINIT)
    return h5chunkit_from_temp(tmp_schunk, b2_schunk)


def export_root(c2_iter: Iterator[os.DirEntry], h5_group: h5py.Group) -> None:
    """Export existing Caterva2 root/directory iterator `c2_iter` into
    open HDF5 group `h5_group`.
    """
    for entry in c2_iter:
        if entry.is_dir(follow_symlinks=True):
            with os.scandir(entry) as c2i:
                h5g = h5_group.create_group(entry.name)
                export_root(c2i, h5g)
        else:
            export_leaf(entry, h5_group)


def export(cat2_path, hdf5_path):
    """Export Caterva2 root at `cat2_path` to new HDF5 file in `hdf5_path`."""
    with os.scandir(cat2_path) as c2i, h5py.File(hdf5_path, 'x') as h5f:
        # keeps directory open, create, fail if existing
        export_root(c2i, h5f)

def main():
    try:
        _, cat2_path, hdf5_path = sys.argv
    except ValueError:
        def eprint(*args):
            print(*args, file=sys.stderr)
        prog = os.path.basename(sys.argv[0])
        eprint(f"Usage: {prog} CATERVA2_ROOT HDF5_FILE")
        eprint("Export the existing CATERVA2_ROOT directory "
               "as a hierarchy in the new HDF5_FILE.")
        eprint('\n'.join(__doc__.splitlines()[1:]))
        sys.exit(1)
    else:
        export(cat2_path, hdf5_path)


if __name__ == '__main__':
    main()



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

"""Export datasets in an HDF5 file to a Caterva2 root.

The HDF5 file is opened in read-only mode, and the Caterva2 root directory is
created anew (it must not exist).  For each group in the HDF5 file hierarchy,
a directory with the same path name is created under the Caterva2 root.  For
each dataset, a Blosc2 dataset with the same path name (plus the ``.b2nd``
extension) is created under the Caterva2 root.

The only datasets supported currently are those which can be converted to
NumPy arrays.  Dataset attributes are supported as Blosc2 vlmeta entries if
they can be serialized with msgpack (with NumPy and empty values first
translated into native Python).  Group attributes are not supported yet.
Moreover, for the moment datasets get compressed with default Blosc2
parameters.

Warning: For the moment, the data in each dataset is read and decompressed
into memory in its entirety.

Datasets or attributes which are unsupported or fail to be converted are
simply reported and skipped, and they do not cause the program to fail.
"""

import logging
import os
import pathlib
import sys
from collections.abc import Callable, Iterator

import blosc2
import h5py

from .. import hdf5


def create_directory(name: str, node: h5py.Group,
                     c2_root: pathlib.Path) -> None:
    if len(node.attrs.keys()) > 0:
        logging.warning(f"Exporting group attributes "
                        f"is not supported yet: {name!r}")

    path = c2_root / name
    try:
        path.mkdir()  # parent should exist, not itself
    except OSError as ose:
        logging.error(f"Failed to create directory "
                      f"for node: {name!r} -> {ose!r}")
        return
    logging.info(f"Exported group: {name!r} => {str(path)!r}")


def b2mkempty_b2chunkit_from_dataset(node: h5py.Dataset) -> (
        Callable[..., blosc2.NDArray],
        Iterator[bytes]):
    """Get empty Blosc2 array maker and compressed chunk iterator from `node`.

    The first returned value can be called to create an empty Blosc2 array
    with prepared construction arguments and attributes extracted from the
    HDF5 dataset `node`.  By default, it is created without storage, but that
    may be changed by passing additional keyword arguments like ``urlpath``.

    The second returned value is an iterator that yields compressed Blosc2
    chunks compatible with the construction arguments used by the previous
    callable, for the data in `node`.  They may be stored straight away in
    order in a Blosc2 super-chunk without further processing.
    """
    b2_args = hdf5.b2args_from_h5dset(node)
    b2_attrs = hdf5.b2attrs_from_h5dset(
        node,
        attr_ok=lambda n, a: logging.info(
            f"Translated dataset attribute {a!r}: {n.name!r}"),
        attr_err=lambda n, a, e: logging.error(
            f"Failed to translate dataset attribute "
            f"{a!r}: {n.name!r} -> {e!r}"),
    )
    _, b2iterchunks = hdf5.b2chunkers_from_h5dset(node, b2_args)

    def b2_make_empty(**kwds) -> blosc2.NDArray:
        return hdf5.b2empty_from_h5dset(node, b2_args, b2_attrs, **kwds)

    return b2_make_empty, b2iterchunks()


def copy_dataset(name: str, node: h5py.Dataset,
                 c2_root: pathlib.Path) -> None:
    # TODO: handle array / frame / (compressed) file distinctly

    try:
        b2mkempty, b2_chunks = b2mkempty_b2chunkit_from_dataset(node)
    except Exception as e:
        logging.error(f"Failed to translate dataset "
                      f"to Blosc2 ND array: {name!r} -> {e!r}")
        return

    b2_path = c2_root / f'{name}.b2nd'
    try:
        b2_array = b2mkempty(urlpath=b2_path, mode='w')
        b2_schunk = b2_array.schunk
        for (nchunk, chunk) in enumerate(b2_chunks):
            b2_schunk.update_chunk(nchunk, chunk)
    except Exception as e:
        b2_path.unlink(missing_ok=True)
        logging.error(f"Failed to save dataset "
                      f"as Blosc2 ND array: {name!r} -> {e!r}")
        return

    logging.info(f"Exported dataset: {name!r} => {str(b2_path)!r}")


def node_exporter(c2_root: pathlib.Path):
    """Return an HDF5 node item visitor to export to
    existing Caterva2 root at `c2_root`.
    """
    def export_node(name, node):
        if any(d in [os.path.curdir, os.path.pardir] for d in name.split('/')):
            logging.warning(f"Invalid node name, skipping: {name!r}")
            return
        if isinstance(node, h5py.Group):
            do_export_node = create_directory
        elif isinstance(node, h5py.Dataset):
            do_export_node = copy_dataset
        else:
            logging.warning(f"Unsupported node type "
                            f"{type(node).__name__}, skipping: {name!r}")
            return

        do_export_node(name, node, c2_root)
    return export_node


def export_group(h5_group: h5py.File, c2_root: pathlib.Path) -> None:
    """Export open HDF5 group `h5_group` to
    existing Caterva2 root at `c2_root`.
    """
    # TODO: soft & external links (not visited)
    h5_group.visititems(node_exporter(c2_root))


def export(hdf5_path: str, cat2_path: str) -> None:
    """Export HDF5 file in `hdf5_path` to new Caterva2 root at `cat2_path`."""
    with h5py.File(hdf5_path, 'r') as h5f:
        c2r = pathlib.Path(cat2_path).resolve()
        c2r.mkdir(parents=True)

        export_group(h5f, c2r)


def main():
    try:
        _, hdf5_path, cat2_path = sys.argv
    except ValueError:
        def eprint(*args):
            print(*args, file=sys.stderr)
        prog = os.path.basename(sys.argv[0])
        eprint(f"Usage: {prog} HDF5_FILE CATERVA2_ROOT")
        eprint("Export the hierarchy in the existing HDF5_FILE "
               "into the new CATERVA2_ROOT directory.")
        eprint('\n'.join(__doc__.splitlines()[1:]))
        sys.exit(1)
    else:
        export(hdf5_path, cat2_path)


if __name__ == '__main__':
    main()



"""
Configuration for the subscriber only.

TODO Move toml config here.
"""

import re

from caterva2 import utils


def parse_size(size):
    if size is None:
        return None

    units = {
        "": 1,
        "K": 2**10,
        "k": 2**10,
        "M": 2**20,
        "m": 2**20,
        "G": 2**30,
        "g": 2**30,
    }
    m = re.match(r"^([\d\.]+)\s*([a-zA-Z]{0,3})$", str(size).strip())
    number, unit = float(m.group(1)), m.group(2).upper()
    return int(number * units[unit])


conf = utils.get_conf("subscriber", allow_id=True)

urlbase = conf.get(".urlbase", "http://localhost:8002")
login = conf.get(".login", True)
register = conf.get(".register", False)
demo = conf.get(".demo", False)

quota = parse_size(conf.get(".quota"))
maxusers = conf.get(".maxusers")


# Not strictly necessary but useful for documentation
statedir = None
database = None  # <Database> instance
cache = None
personal = None
shared = None
public = None



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

from collections.abc import Callable, Iterator, Mapping

# Requirements
import blosc2
import h5py
import msgpack
import numpy

"""The registered identifier for Blosc2 in HDF5 filters."""
BLOSC2_HDF5_FID = 32026


# Warning: Keep the reference to the returned result.
# Losing the reference to the array may result in a segmentation fault.
def b2_from_h5chunk(h5_dset: h5py.Dataset,
                    chunk_index: int) -> (blosc2.NDArray | blosc2.SChunk):
    h5chunk_info = h5_dset.id.get_chunk_info(chunk_index)
    return blosc2.open(h5_dset.file.filename, mode='r',
                       offset=h5chunk_info.byte_offset)


def h5dset_is_compatible(h5_dset: h5py.Dataset) -> bool:
    """Is the HDF5 dataset compatible with a Blosc2 dataset?"""
    shape = h5_dset.shape
    if shape is None:
        return False  # empty dataspace (``H5S_NULL``)
    if len(shape) > getattr(blosc2, 'MAX_DIM', 8):
        return False  # too many dimensions
    dtype = h5_dset.dtype
    if dtype.ndim != 0 or dtype.fields is not None:
        return False  # array or compound dtype
    if dtype.kind in ['O']:  # other kinds may be missing
        return False
    return True


def b2args_from_h5dset(h5_dset: h5py.Dataset) -> Mapping[str, object]:
    """Get Blosc2 array creation arguments for the given HDF5 dataset.

    Return a mapping which can be applied to array creation calls.

    This may be an expensive operation with Blosc2-compressed datasets, as it
    requires opening the Blosc2 super-chunk in the first HDF5 chunk to read
    its metadata (so as to return the exact same arguments).
    """
    if not h5dset_is_compatible(h5_dset):
        raise TypeError("HDF5 dataset is not compatible with Blosc2")

    b2_args = {
        'chunks': h5_dset.chunks,  # None is ok (let Blosc2 decide)
    }

    if (h5_dset.chunks is None
            or list(h5_dset._filters) != [f'{BLOSC2_HDF5_FID:#d}']
            or h5_dset.id.get_num_chunks() < 1):
        return b2_args

    # Blosc2 is the sole filter, direct chunk copy is possible.
    # Get Blosc2 arguments from the first schunk.
    # HDF5 filter parameters are less reliable than these.
    b2_array = b2_from_h5chunk(h5_dset, 0)
    b2_schunk = getattr(b2_array, 'schunk', b2_array)
    if hasattr(b2_array, 'blocks'):  # rely on cparams blocksize otherwise
        b2_args['blocks'] = b2_array.blocks
    b2_args['cparams'] = b2_schunk.cparams
    b2_args['dparams'] = b2_schunk.dparams

    return b2_args


def _msgpack_h5attr(obj):
    if isinstance(obj, tuple):  # ad hoc Blosc2 tuple handling
        return ['__tuple__', *obj]

    if isinstance(obj, h5py.Empty):
        if obj.dtype.kind not in ['S', 'U']:  # not strings
            # This drops object type but is probably less dangerous
            # than using an actual zero value.
            return None
        obj = obj.dtype.type()  # use empty value of that type

    if isinstance(obj, (numpy.generic, numpy.ndarray)):
        return obj.tolist()

    return obj


def b2attrs_from_h5dset(
        h5_dset: h5py.Dataset,
        attr_ok: Callable[[h5py.Dataset, str], None] | None = None,
        attr_err: Callable[[h5py.Dataset, str, Exception], None] | None = None) -> (
            Mapping[str, object]):
    """Get msgpack-encoded attributes from the given HDF5 dataset.

    NumPy and empty attribute values are first translated into native Python
    values.

    If given, call `attr_ok` or `attr_err` on attribute translation success or
    error, respectively.
    """
    b2_attrs = {}
    for (aname, avalue) in h5_dset.attrs.items():
        try:
            # This workaround allows NumPy objects
            # and converts them to similar native Python objects
            # which can be encoded by plain msgpack.
            # Of course, some typing information is lost in the process,
            # but the result is portable.
            pvalue = msgpack.packb(avalue, default=_msgpack_h5attr,
                                   strict_types=True)
        except Exception as e:
            if attr_err:
                attr_err(h5_dset, aname, e)
        else:
            b2_attrs[aname] = pvalue
            if attr_ok:
                attr_ok(h5_dset, aname)
    return b2_attrs


def _b2maker_from_h5dset(b2make: Callable[..., blosc2.NDArray]):
    """Get a factory to create a Blosc2 array compatible with an HDF dataset.

    The result may be called with the dataset and optional Blosc2 creation
    arguments and msgpack-encoded attributes (which will be added to the
    array's variable-length metadata), plus other keyword arguments.  The
    dataset-compatible Blosc2 array will be created using the `b2make`
    callable.

    If the aforementioned optional parameters are not given, they will be
    extracted anew from the HDF5 dataset (which may be an expensive operation
    depending on the case).
    """
    def _b2make_from_h5dset(h5_dset: h5py.Dataset,
                            b2_args=None, b2_attrs=None,
                            **kwds) -> blosc2.NDArray:
        b2_args = (b2_args if b2_args is not None
                   else b2args_from_h5dset(h5_dset))
        b2_attrs = (b2_attrs if b2_attrs is not None
                    else b2attrs_from_h5dset(h5_dset))

        b2_array = b2make(shape=h5_dset.shape, dtype=h5_dset.dtype,
                          **(b2_args | kwds))
        b2_vlmeta = b2_array.schunk.vlmeta
        for (aname, avalue) in b2_attrs.items():
            b2_vlmeta.set_vlmeta(aname, avalue, typesize=1)  # non-numeric
        return b2_array
    return _b2make_from_h5dset


b2empty_from_h5dset = _b2maker_from_h5dset(blosc2.empty)
b2uninit_from_h5dset = _b2maker_from_h5dset(blosc2.uninit)


def b2chunkers_from_h5dset(h5_dset: h5py.Dataset, b2_args=None) -> (
        Callable[[int], bytes],
        Callable[[], Iterator[bytes]]):
    """Get by-index and iterator chunkers for the given dataset.

    The first returned value (``getchunk``) can be called with an integer
    index to get the compressed Blosc2 chunk with that index.  An `IndexError`
    is raised if the index is out of bounds.

    The second returned value (``iterchunks``) can be called to get an
    iterator that yields compressed Blosc2 chunks in order.

    Chunks will be compatible with Blosc2 array construction arguments, either
    as provided by a non-empty mapping `b2_args`, or otherwise inferred from
    the dataset.  The former may be useful if you have already called
    `b2args_from_h5dset()` on the dataset.  The chunks may be stored straight
    away in order in a Blosc2 super-chunk without further processing.

    Some chunkers may hold a sizable amount of memory (especially for datasets
    with big chunks, or big non-chunked datasets).
    """
    b2_args = b2_args or b2args_from_h5dset(h5_dset)

    if b2_args['chunks'] is None:
        b2chunkers_from_dataset = b2chunkers_from_nonchunked
    elif 'cparams' in b2_args:
        b2chunkers_from_dataset = b2chunkers_from_blosc2
    else:
        b2chunkers_from_dataset = b2chunkers_from_chunked

    return b2chunkers_from_dataset(h5_dset, b2_args)


def b2chunkers_from_blosc2(h5_dset: h5py.Dataset, b2_args: Mapping) -> (
        Callable[[int], bytes],
        Callable[[], Iterator[bytes]]):
    # Blosc2-compressed dataset, just pass chunks as they are.
    # Support both Blosc2 arrays and frames as HDF5 chunks.
    def b2getchunk_blosc2(nchunk: int) -> bytes:
        if not (0 <= nchunk < h5_dset.id.get_num_chunks()):
            raise IndexError(nchunk)
        return _b2getchunk_nchunk(nchunk)

    def _b2getchunk_nchunk(nchunk: int) -> bytes:
        b2_array = b2_from_h5chunk(h5_dset, nchunk)
        b2_schunk = getattr(b2_array, 'schunk', b2_array)
        # TODO: check if schunk is compatible with creation arguments
        if b2_schunk.nchunks < 1:
            raise OSError(f"chunk #{nchunk} of HDF5 node {h5_dset.name!r} "
                          f"contains Blosc2 super-chunk with no chunks")
        if b2_schunk.nchunks > 1:
            # TODO: warn, check shape, re-compress as single chunk
            raise NotImplementedError(
                f"chunk #{nchunk} of HDF5 node {h5_dset.name!r} "
                f"contains Blosc2 super-chunk with several chunks")
        return b2_schunk.get_chunk(0)

    def b2iterchunks_blosc2() -> Iterator[bytes]:
        for nchunk in range(h5_dset.id.get_num_chunks()):
            yield _b2getchunk_nchunk(nchunk)

    return b2getchunk_blosc2, b2iterchunks_blosc2


def b2chunkers_from_nonchunked(h5_dset: h5py.Dataset, b2_args: Mapping) -> (
        Callable[[int], bytes],
        Callable[[], Iterator[bytes]]):
    # Contiguous or compact dataset,
    # slurp into Blosc2 array and get chunks from it.
    # Hopefully the data is small enough to be loaded into memory.
    b2_array = blosc2.asarray(
        numpy.asanyarray(h5_dset[()]),  # ok for arrays & scalars
        **b2_args
    )

    def b2getchunk_nonchunked(nchunk: int) -> bytes:
        if not (0 <= nchunk < b2_array.schunk.nchunks):
            raise IndexError(nchunk)
        return _b2getchunk_nchunk(nchunk)

    def _b2getchunk_nchunk(nchunk: int) -> bytes:
        return b2_array.schunk.get_chunk(nchunk)

    def b2iterchunks_nonchunked() -> Iterator[bytes]:
        for chunk_info in b2_array.schunk.iterchunks_info():
            yield _b2getchunk_nchunk(chunk_info.nchunk)

    return b2getchunk_nonchunked, b2iterchunks_nonchunked


def b2chunkers_from_chunked(h5_dset: h5py.Dataset, b2_args: Mapping) -> (
        Callable[[int], bytes],
        Callable[[], Iterator[bytes]]):
    # Non-Blosc2 chunked dataset,
    # load each HDF5 chunk into chunk 0 of compatible Blosc2 array,
    # then get the resulting compressed chunk.
    # Thus, only one chunk worth of data is kept in memory.
    assert h5_dset.chunks == b2_args['chunks']
    b2_array = blosc2.empty(
        shape=h5_dset.chunks, dtype=h5_dset.dtype,  # array shape is chunkshape
        **b2_args
    )

    def b2getchunk_chunked(nchunk: int) -> bytes:
        if not (0 <= nchunk < h5_dset.id.get_num_chunks()):
            raise IndexError(nchunk)
        chunk_start = h5_dset.id.get_chunk_info(nchunk).chunk_offset
        chunk_slice = tuple(slice(cst, cst + csz, 1)
                            for (cst, csz) in zip(chunk_start, h5_dset.chunks, strict=False))
        return _b2getchunk_slice(chunk_slice)

    def _b2getchunk_slice(chunk_slice) -> bytes:
        chunk_array = numpy.asanyarray(h5_dset[chunk_slice])
        # Always place at the beginning so that it fits in chunk 0.
        b2_slice = tuple(slice(0, n, 1) for n in chunk_array.shape)
        b2_array[b2_slice] = chunk_array
        return b2_array.schunk.get_chunk(0)

    def b2iterchunks_chunked() -> Iterator[bytes]:
        for chunk_slice in h5_dset.iter_chunks():
            yield _b2getchunk_slice(chunk_slice)

    return b2getchunk_chunked, b2iterchunks_chunked



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

"""
Script for exporting a series of images into a Blosc2 HDF5 file using the Grok codec.
"""

import blosc2_grok
import h5py
import hdf5plugin
import numpy as np
from PIL import Image
import os

import blosc2

if __name__ == '__main__':
    # Define the compression and decompression parameters. Disable the filters and the
    # splitmode, because these don't work with the codec.
    cparams = {
        'codec': blosc2.Codec.GROK,
        'nthreads': 1,
        'filters': [],
        'splitmode': blosc2.SplitMode.NEVER_SPLIT,
    }
    # Set the parameters that will be used by grok
    kwargs = {
        'cod_format': blosc2_grok.GrkFileFmt.GRK_FMT_JP2,
        'num_threads': 1,    # this does not have any effect (grok should work in multithreading mode)
        'quality_mode': "rates",
    }

    # Compression params for identifying the codec. In the future, one should be able to
    # specify the grok plugin (and its parameters) here.
    b2params = hdf5plugin.Blosc2()

    # Open the PNG datasets
    dir_path = 'numbers'
    # List of PNG files (in order)
    png_files = [
        "zero.png",
        "one.png",
        "two.png",
        "three.png",
        "four.png",
        "five.png",
        "six.png",
        "seven.png",
        "eight.png",
        "nine.png",
    ]
    images_color = []
    images_gray = []
    for png_file in png_files:
        # Open the image file
        img_color = Image.open(os.path.join(dir_path, png_file))
        images_color.append(np.array(img_color))
        # Convert the image to grayscale
        img_gray = img_color.convert('L')
        images_gray.append(np.array(img_gray))
    dset_color = np.array(images_color)
    dset_gray = np.array(images_gray)

    # Open the output file for color and gray images
    fname = f'numbers-jpeg2000.h5'
    print(f"output file: {fname}")
    fout = h5py.File(fname, 'w')

    for cratio in (2, 5, 10, 20):
        print(f"Compressing with cratio={cratio}x ...")
        kwargs['quality_layers'] = np.array([cratio], dtype=np.float64)
        blosc2_grok.set_params_defaults(**kwargs)
        group = fout.create_group(f'/cratio_{cratio}x')
        # Store the color and grey images
        for dset in [dset_color, dset_gray]:
            chunks = (1,) + dset.shape[1:]
            name = 'numbers_color' if dset is dset_color else 'numbers_gray'
            disk_dset = group.create_dataset(name, shape=dset.shape, dtype=dset.dtype,
                                             chunks=chunks, **b2params)
            disk_dset.attrs['contenttype'] = 'tomography'
            for i in range(dset.shape[0]):
                im = dset[i:i+1]
                b2im = blosc2.asarray(im, chunks=im.shape, blocks=im.shape, cparams=cparams)
                # Write to disk.
                # Due to leading 1 dim, color images have 4 dimensions, gray images have 3 dimensions
                offset = (i, 0, 0, 0) if dset is dset_color else (i, 0, 0)
                disk_dset.id.write_direct_chunk(offset, b2im.schunk.to_cframe())

    fout.close()



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2025 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

# Example showing how to chain lazy expressions and retrieve just a
# (compressed) slice of the result.  A temporary array is created
# and uploaded to the server, and then removed and the end.

from time import time

import blosc2
import numpy as np

import caterva2 as cat2

# Open a client to the local server
# client = cat2.Client("http://localhost:8002", ("user@example.com", "foobar11"))
# Open a client to the Cat2Cloud server
client = cat2.Client("https://cat2.cloud/demo", ("user@example.com", "foobar11"))

a = blosc2.linspace(0, 100, 100_000_000, dtype="float32", shape=(10000, 10000), urlpath="a.b2nd", mode="w")
t0 = time()
apath = client.upload("a.b2nd", "@public/a.b2nd")
ra = client.get(apath)
t1 = time() - t0
print(
    f"Time for uploading data (HTTP): {t1:.3f}s"
    f" - file size: {a.schunk.nbytes / 2**20:.2f} MB"
    f" - speed: {a.schunk.nbytes / 2**20 / t1:.2f} MB/s"
)

la1 = client.lazyexpr("la1", "a + 1", {"a": apath})
la2 = client.lazyexpr("la2", "la1 + 1", {"la1": la1})
rla2 = client.get(la2)

# Compute a slice of the remote array
t0 = time()
result = rla2.slice(slice(5, 9))
print(
    f"Time for computing and downloading data (compressed): {time() - t0:.3f}s"
    f" - data size: {result.schunk.nbytes / 2**10:.2f} KB"
)

np.testing.assert_allclose((a + 2)[5:9], result[:], rtol=1e-6)

# Cleanup
client.remove(apath)



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

"""Subscriber database.

Used for user authentication (for the moment), based on FastAPI Users example:
https://fastapi-users.github.io/fastapi-users/latest/configuration/full-example/

The database should only be used if a (non-empty) secret token for the
management of users is set in the 'CATERVA2_SECRET' environment variable.

The database will be stored in SQLite format inside the state directory
given to `create_db_and_tables()`.
"""

from collections.abc import AsyncGenerator
from pathlib import Path

from fastapi import Depends
from fastapi_users.db import SQLAlchemyBaseUserTableUUID, SQLAlchemyUserDatabase
from sqlalchemy.ext.asyncio import AsyncSession, async_sessionmaker, create_async_engine
from sqlalchemy.orm import DeclarativeBase


class Base(DeclarativeBase):
    pass


class User(SQLAlchemyBaseUserTableUUID, Base):
    pass


engine = None
async_session_maker = None


async def create_db_and_tables(statedir: Path):
    global engine, async_session_maker  # keep global as in original example
    engine = create_async_engine(f'sqlite+aiosqlite:///{statedir}/db.sqlite')
    async_session_maker = async_sessionmaker(engine, expire_on_commit=False)

    async with engine.begin() as conn:
        await conn.run_sync(Base.metadata.create_all)


async def get_async_session() -> AsyncGenerator[AsyncSession, None]:
    async with async_session_maker() as session:
        yield session


async def get_user_db(session: AsyncSession = Depends(get_async_session)):
    yield SQLAlchemyUserDatabase(session, User)



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

"""Caterva2 - On demand access to remote Blosc2 data repositories"""

from .client import BasicAuth, Client, Dataset, File, Root, sub_urlbase_default

__version__ = "2025.5.2"
"""The version in use of the Caterva2 package."""

__all__ = [
    "BasicAuth",
    "Client",
    "Root",
    "File",
    "Dataset",
    "sub_urlbase_default",
]
"""List of symbols exported by the Caterva2 package."""



import sys

import pytest


def main(verbose=False, args=None):
    """Run the test suite.

    Parameters
    ----------
    verbose : bool
        If True, run the tests in verbose mode.
    args : list[string]
        Arguments to pass to pytest execution.

    Returns
    -------
    int
        Exit code of the test suite.
    """
    args = [] if args is None else args.copy()
    if verbose:
        args.append("-v")
    # This only happens when using this script mechanism to run tests,
    # using ``pytest --pyargs caterva2.tests`` is not affected.
    args.append("-Wignore::pytest.PytestAssertRewriteWarning")
    args.append("--pyargs")
    args.append(__package__)
    return pytest.main(args)


if __name__ == '__main__':
    main(args=sys.argv)



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

import datetime

# Requirements
import blosc2
import pydantic


class CParams(pydantic.BaseModel, extra="allow"):
    codec: blosc2.Codec
    codec_meta: int
    clevel: int
    filters: list[blosc2.Filter]
    filters_meta: list[int]
    typesize: int
    blocksize: int
    nthreads: int
    splitmode: blosc2.SplitMode
    tuner: blosc2.Tuner
    use_dict: bool


class SChunk(pydantic.BaseModel, extra="allow"):
    cbytes: int
    chunkshape: int
    chunksize: int
    contiguous: bool
    cparams: CParams
    cratio: float
    #   dparams
    #   meta
    nbytes: int
    urlpath: str
    vlmeta: dict = {}
    nchunks: int
    mtime: datetime.datetime | None = None


class Metadata(pydantic.BaseModel):
    shape: tuple
    chunks: tuple
    blocks: tuple
    dtype: str
    schunk: SChunk
    mtime: datetime.datetime | None


class LazyArray(pydantic.BaseModel):
    shape: tuple | None
    dtype: str
    expression: str
    operands: dict[str, str | None]
    mtime: datetime.datetime | None


class NewLazyExpr(pydantic.BaseModel):
    name: str
    expression: str
    operands: dict[str, str]
    compute: bool


class MoveCopyPayload(pydantic.BaseModel):
    src: str
    dst: str


class AddUserPayload(pydantic.BaseModel):
    username: str
    password: str | None
    superuser: bool


class File(pydantic.BaseModel):
    mtime: float
    size: int


class Root(pydantic.BaseModel):
    name: str
    http: str
    subscribed: bool | None = None  # Used only by the subscriber program


class Broker(pydantic.BaseModel):
    roots: dict[str, Root]


class Publisher(pydantic.BaseModel):
    etags: dict[str, str]


class Subscriber(pydantic.BaseModel):
    roots: dict[str, Root]
    etags: dict[str, str]
